(* this module is for loading our supported file types *)

open Datapoint
open Util

let words = Str.split (Str.regexp "[ \t]+")

let rec last = function
    [] -> internal_error "last: []"
  | [x] -> x
  | (_::xs) -> last xs

let getFeatureNum linenum s =
  match Str.split (Str.regexp ":") s with
      [f;_] -> int_of_string f
    | (_::f::_) -> int_of_string f
    | _ -> failwith ("could not identify feature number on line " ^ string_of_int linenum ^ ", '" ^ s ^ "'")

let string_sub str low len =
  String.sub str low (min len (String.length str - low))

let getFileInfoSVM filepath =
  let c = open_in filepath in
  let ln = ref 0 in
  let lc = ref 0 in
  let lf = ref (-1) in
  let hf = ref 0 in
  let hasClasses = ref true in
    (try
       while (true) do
	 let _ = ln := !ln+1 in
	 let s = input_line c in
	   if String.length s == 0 
	   then ()
	   else if String.get s 0 == '#'
	   then ()
	   else 
	     begin
               let wrds = words s in
               let x  = List.hd wrds in
               let xs = List.tl wrds in
	       let dps = if String.contains x ':' then (x::xs) else xs in
	       let flo = getFeatureNum (!ln) (List.hd dps) in
	       let fhi = getFeatureNum (!ln) (last dps) in
		 (* we assume the features are in order *)
		 if (String.contains x ':') then hasClasses := false;
		 lc := !lc + 1;
		 if (!lf < 0) || (flo < !lf) then lf := flo;
		 if (fhi > !hf) then hf := fhi;
	     end
       done;
     with
	 End_of_file -> ()
       | e -> raise e);
    close_in c;
    (!hasClasses, !lc, !lf, !hf)

let getFileInfoSeparated regexp filepath =
  let c = open_in filepath in
  let ln = ref 0 in
  let lc = ref 0 in
  let hf = ref 0 in
    (try
       while (true) do
	 let _ = ln := !ln+1 in
	 let s = input_line c in
	   if String.length s == 0 
	   then ()
	   else if String.get s 0 == '#'
	   then ()
	   else 
	     begin
	       let fhi = List.length (Str.split regexp s) in
		 lc := !lc + 1;
		 if (fhi > !hf) then hf := fhi;
	     end
       done;
     with
	 End_of_file -> ()
       | e -> raise e);
    close_in c;
    (false, !lc, 1, !hf)

let addToDataSVM 
  (filepath : string)
  (hasclasses : bool)
  (hasrank : bool)
  (setClass : int -> float -> unit)      (* data point # -> class, only called when hasclasses is true *)
  (setRank : int -> int -> unit)         (* data point # -> qid, only called with hasrank is true *)
  (addFV : int -> int -> string -> unit) (* data point # -> feature number (unchanged) -> feature value *)
  : unit =
  let c = open_in filepath in
  let lc = ref 0 in
  let ln = ref 0 in
    (try
      (while (true) do
	 let _ = ln := !ln+1 in
	 let s = input_line c in
	   if String.length s == 0 
	   then ()
	   else if String.get s 0 == '#'
	   then ()
	   else 
	     begin
	       let l = words s in
		 if hasclasses then setClass (!lc) (float_of_string (List.hd l));
		 List.iter
		   (function x ->
		      if string_sub x 0 4 = "qid:"
		      then 
			if hasrank then setRank (!lc) (int_of_string (string_sub x 4 (String.length x - 4)))
			else failwith ("not expecting qid: on line " ^ string_of_int (!ln))
		      else if string_sub x 0 7 = "string:"
		      then 
			let i = String.index_from x 7 ':' in
			  addFV (!lc) (int_of_string (string_sub x 7 (i - 7))) (string_sub x i (String.length x - i))
		      else if string_sub x 0 5 = "tree:"
		      then 
			let i = String.index_from x 5 ':' in
			  addFV (!lc) (int_of_string (string_sub x 5 (i - 5))) (string_sub x i (String.length x - i))
		      else if string_sub x 0 9 = "discrete:"
		      then 
			let i = String.index_from x 9 ':' in
			  addFV (!lc) (int_of_string (string_sub x 9 (i - 9))) (string_sub x i (String.length x - i))
		      else 
			match Str.split (Str.regexp ":") x with
			    [f;v] -> addFV (!lc) (int_of_string f) v
			  | _ -> failwith ("error on line " ^ string_of_int (!ln)))
		   (if hasclasses then List.tl l else l);
		 lc := !lc + 1;
	     end
       done)
    with
	End_of_file -> ()
      | e -> raise e);
 close_in c;
 ()


let addToDataSeparated 
  (regexp : Str.regexp)
  (filepath : string)
  (addFV : int -> int -> string -> unit)  (* data point # -> feature number (unchanged) -> feature value *)
  : unit =
  let c = open_in filepath in
  let lc = ref 0 in
  let ln = ref 0 in
    (try
      (while (true) do
	 let _ = ln := !ln+1 in
	 let s = input_line c in
	   if String.length s == 0 
	   then ()
	   else if String.get s 0 == '#'
	   then ()
	   else 
	     begin
	       let l = words s in
	       let i = ref 0 in
		 List.iter (function s ->
			      addFV (!lc) (!i) s;
			      i := !i + 1
			   ) l;
		 lc := !lc + 1;
	     end
       done)
    with
	End_of_file -> ()
      | e -> raise e);
 close_in c;
 ()

(* 
format:
  x0*x0  x0*x1  x0*x2  ..... x0*xn
         x1*x1  x1*x2  ..... x1*xn
                x2*x2  ..... x2*xn
                       ..... .....
                             xn*xn

or full matrix; output matrix is of the upper triangular format, so
position col:row is only valid when col >= row and is stores in m
at position m.(row).(col-row)
*)
let loadKernelMatrix filepath =
  let c = open_in filepath in
  let l = List.map float_of_string (words (input_line c)) in
  let n = List.length l in
  let m = Array.init n (fun i -> Array.make (n-i) 0.) in
  let _ = List.fold_left (fun i v -> m.(0).(i) <- v; i+1) 0 l in
    for row=1 to n-1 do
      let l = Array.of_list (words (input_line c)) in
        if Array.length l == n then
          for col=row to n-1 do
            m.(row).(col-row) <- float_of_string l.(col);
          done
        else 
          for col=row to n-1 do
            m.(row).(col-row) <- float_of_string l.(col-row);
          done
    done;
    close_in c;
    m
