(* Data types for SVMsequel *)

(* we hash (or otherwise store) strings in a vocabulary to speed
   up, eg, dp string computations and to allow for strings in
   discrete features *)
type fstring = int (* identifier *)

(* for dense and sparse vectors of features *)
type denseVector = float array
type sparseVector = int array * float array

(* for strings and trees with k calculated with DP *)
type dpString = fstring array

(* for strings and trees with k calculated with suffix trees *)
type stString = SuffixTreeKernel.streeData

(* for discrete values *)
type discrete = fstring

(* an fvclump is a single bunch of the above calculated with the same
   parameters *)
type fvClump =
    DV of denseVector
  | SV of sparseVector
  | DP of dpString
  | ST of stString
  | Disc of discrete
  | Mat of int  (* kernel matrix given, int is position *)

(* a data set parameterization consists of the different parameters
   for the different clumps that appear in its values *)
type kernelWeight = float
type realKernel =
    LinearKernel
  | PolynomialKernel of float * float * float (* a,b,d, (a+buv)^d *)
  | RbfKernel of float (* gamma *)
  | SigKernel of float * float (* a,b, tansig (a+buv) *)
  | IdKernel of float (* beta *)

type fvParams =
    DVP of kernelWeight * realKernel
  | SVP of kernelWeight * realKernel * int (* highest feature number *)
  | DPP of kernelWeight * float * int (* lambda value and maxn *)
  | STP of kernelWeight * SuffixTreeKernel.weightType
  | DiscP of kernelWeight * int * float (* cardinality of possibilities, beta *)
  | MatP  of kernelWeight * (float array array)
  | FVParamPlaceHolder

(* training data is a bunch of data together with parameters *)
type dataSet =
    { mutable trainingParams : fvParams array;  (* the fvParams and fvClumps arrays must be the same length *)
      mutable datapoints     : fvClump array array;  (* datapoints.(i) is a single training instance, datapoints.(i).(j) is the jth clump of the ith data point *)
      mutable dataclasses    : float array;  (* for training data *)
      mutable datanormalize  : float array array option; (* only exists for dense and sparse data *)
      mutable rankingPoints  : int array option; (* NONE if not ranking, SOME a gives an
					            array where a[i] is the beginning of
					            ranking group i in trainingData *)

      mutable crossValSplit : (int array * int) option;

      mutable dataHasClasses : bool
    }

type multiType = OVA of int array | OVN of int array | AVA of int array

type modelData =
    { modelParams    : fvParams array;  (* copied from dataSet *)
      supportVectors : fvClump array array; (* also copied, but only a_i > 0 *)
      multipliers : float array array; (* this is a_i * y_i, or whatever is
					  appropriate for the method of
					  multi-class that we do *)
      mutable modelnormalize : float array array option;
      mutable modelBias : float array; (* if so desired, ow [||] *)
      mutable multiclass : multiType option;
      mutable binarySigmoid : (float*float) option; (* for binary classification *)
      mutable multiSigmoid : float array option; (* for multi-class classification *)
    }
