(* all this module does is compute dot products *)

open Datapoint
open Array
open Util
open Bigarray

(*
type int_arr   = (int, int_elt, c_layout) Array1.t
type float_arr = (float, float32_elt, c_layout) Array1.t

external sparse_dot_c : int_arr -> float_arr -> int_arr -> float_arr -> float = "sparse_dot"

let (xf : int_arr  ) = Array1.of_array int     c_layout [| 1 ; 2 ; 5 |]
let (yf : float_arr) = Array1.of_array float32 c_layout [| 1.; 2.; 1.|]

let _ = Printf.printf "%g\n" (sparse_dot_c xf yf xf yf)
*)

module H = IntHashtbl

let maxCacheSize = ref 1000

let (kernelCache : float H.t H.t) = H.create 10
let kernelCacheSize = ref 0
let clearCache () = H.clear kernelCache ; kernelCacheSize := 0

let checkCache xid yid =
  let lo = min xid yid in let hi = max xid yid in
    try
      Some (H.find (H.find kernelCache lo) hi)
    with
        Not_found -> None
      | e -> raise e

let addToCache xid yid v =
  if !kernelCacheSize > !maxCacheSize then () else
    let lo = min xid yid in let hi = max xid yid in
      (try 
         H.add (H.find kernelCache lo) hi v
       with
           Not_found ->
             ( let h = H.create 10 in
                 H.add h hi v ;
                 H.add kernelCache lo h )
         | e -> raise e) ;
      kernelCacheSize := !kernelCacheSize+1


(*
let (kernelCache : (int*int, float) H.t) = H.create 50
let kernelCacheSize = ref 0

let clearCache () = H.clear kernelCache ; kernelCacheSize := 0

let checkCache xid yid =
  let p = if xid <= yid then (xid,yid) else (yid,xid) in
    try
      Some (H.find kernelCache p)
    with
        Not_found -> None
      | e -> raise e

let addToCache xid yid v =
  if !kernelCacheSize >= !maxCacheSize then ()
  else begin
    let p = if xid <= yid then (xid,yid) else (yid,xid) in
      kernelCacheSize := !kernelCacheSize + 1;
      H.add kernelCache p v
  end
*)

let denseKernel k x y =
  let l = length x in
  let dotValue x y = 
    let s = ref 0.0 in
      (match k with
          IdKernel _ ->
            for i = 0 to l-1 do
	      s := !s +. sqrt(x.(i) *. y.(i));
            done;
        | _ ->
            for i = 0 to l-1 do
	      s := !s +. x.(i) *. y.(i);
            done);
      !s in
  let xdoty = dotValue x y in
    match k with
	LinearKernel -> xdoty
      | PolynomialKernel (a,b,d) -> (a +. b *. xdoty) ** d
      | RbfKernel g -> exp (g *. (2. *. xdoty -. dotValue x x -. dotValue y y))
      | SigKernel (a,b) -> tanh (a +. b *. xdoty)
      | IdKernel b -> 
          let v = acos xdoty in
            (4.*.3.14159265*.b) ** (-0.5 *. (float_of_int l -. 1.)) *.
            exp (0. -. v *. v /. b)

let sparseKernel k (highestFeature:int) (xf,xv) (yf,yv) =  (* highestFeature only used for ID kernel *)
  let lx = length xf in
  let ly = length yf in
  let dotValue xf xv yf yv =
    let s = ref 0.0 in
    let i = ref 0 in
    let j = ref 0 in
      (match k with
           IdKernel _ ->
             while ((!i < lx) && (!j < ly) && (xf.(!i) >= 0) && (yf.(!j) >= 0)) do
	       if xf.(!i) == yf.(!j) 
	       then ( 
		 s := !s +. sqrt(xv.(!i) *. yv.(!j)) ;
		 i := !i + 1;
		 j := !j + 1 )
	       else if xf.(!i) < yf.(!j) 
	       then i := !i + 1
	       else j := !j + 1
             done

         | _ -> 
             while ((!i < lx) && (!j < ly) && (xf.(!i) >= 0) && (yf.(!j) >= 0)) do
	       if xf.(!i) == yf.(!j) 
	       then ( 
		 s := !s +. xv.(!i) *. yv.(!j) ;
	         i := !i + 1;
	         j := !j + 1 )
	       else if xf.(!i) < yf.(!j) 
	       then i := !i + 1
	       else j := !j + 1
             done);
      !s in
  let dotSelf f a =
    let s = ref 0.0 in
    let l = length a in
      (try
         for i = 0 to l-1 do
	   if f.(i) < 0 then raise End_of_file;
	   s := !s +. a.(i) *. a.(i)
         done
       with
           End_of_file -> ()
         | e -> raise e);
      !s in
  let xdoty = dotValue xf xv yf yv in
    match k with
	LinearKernel -> xdoty
      | PolynomialKernel (a,b,d) -> (a +. b *. xdoty) ** d
      | RbfKernel g -> exp (g *. (2. *. xdoty -. dotSelf xf xv -. dotSelf yf yv))
      | SigKernel (a,b) -> tanh (a +. b *. xdoty)
      | IdKernel b -> 
          let v = acos xdoty in
            (4.*.3.14159265*.b) ** (-0.5 *. (float_of_int highestFeature -. 1.)) *.
            exp (0. -. v *. v /. b)


let dpStringKernel lambda maxn s t = 
  let ls = length s in let lt = length t in
  let kprime = 
    init maxn (function n -> init ls (function i -> make lt 0.)) in

  (* initialize K'_0 *)
    for i = 0 to ls-1 do
      for j = 0 to lt-1 do
        kprime.(0).(i).(j) <- 1.;
      done;
    done;
    
  (* recurse for K' *)
    for n = 1 to maxn-1 do
      for i = 1 to ls-1 do
        let x = s.(i) in
        for j = 1 to lt-1 do
          let oldV = kprime.(n).(i-1).(j) in
          let s    = ref 0. in
            for k = 0 to lt-1 do
              if t.(k) == x
              then s := !s +. kprime.(n-1).(i).(k) *. lambda ** (float_of_int (lt - k + 2));
            done;
            kprime.(n).(i).(j) <- !s +. oldV;
        done;
      done;
    done;

  (* compute K_maxn(s, t) *)
    let kval = ref 0. in

    for i=0 to ls-1 do
      let x = s.(i) in
        for j=0 to lt-1 do
          if t.(j) == x
          then kval := !kval +. kprime.(maxn-1).(i).(j) *. lambda *. lambda;
        done;
    done;

    !kval


let stStringKernel f x y = 
  SuffixTreeKernel.st_string_kernel 
    (SuffixTreeKernel.weightTypeToFn f)
    x.SuffixTreeKernel.tree_ref
    x.SuffixTreeKernel.tree_kvc
    y.SuffixTreeKernel.tree_ref.SuffixTree.tree_string

let discreteKernel sz beta x y = 
  if x == y then 1. else
    let ev = exp (0. -. beta *. float_of_int sz) in
      (1. -. ev) /. (1. +. (float_of_int sz -. 1.) *. ev)

let matrixKernel m x y =
  if x <= y then m.(x).(y) else m.(y).(x)

let dotClumps2
    (norm    : float array array option)
    (clumpsX : fvClump array) 
    (clumpsY : fvClump array) 
    (params  : fvParams array) : float =
  let s = ref 0.0 in
  let normFactor i = match norm with None -> 1. | Some a -> a.(i).(0) in
  let l = length clumpsX in
    for i = 0 to l-1 do
      match (clumpsX.(i), clumpsY.(i), params.(i)) with
	  (DV   x, DV   y, DVP (  w,k)  ) -> s := !s +. w *. denseKernel k x y;
	| (SV   x, SV   y, SVP   (w,k,h)) -> s := !s +. w *. sparseKernel k h x y
	| (DP   x, DP   y, DPP   (w,l,n)) -> s := !s +. w *. dpStringKernel l n x y /. normFactor i
	| (ST   x, ST   y, STP   (w,f)  ) -> s := !s +. w *. stStringKernel f x y   /. normFactor i
	| (Disc x, Disc y, DiscP (w,a,b)) -> s := !s +. w *. discreteKernel a b x y
	| (Mat  x, Mat  y, MatP  (w,m)  ) -> s := !s +. w *. matrixKernel m x y
	| _ -> failwith ("Dot.dotClumps: invalid combination of clumps and params; incompatible data")
    done;
    !s


let dotClumps 
    (norm    : float array array option)
    (clumpsX : fvClump array) 
    (clumpsY : fvClump array) 
    (xid : int)  (* these are for caching, which we don't do right now, <0 means don't cache *)
    (yid : int)
    (params  : fvParams array) : float =
  if xid >= 0 && yid >= 0 then
    match checkCache xid yid with
        Some v -> v
      | None   -> 
          let v = dotClumps2 norm clumpsX clumpsY params in
            ( addToCache xid yid v ; v )
  else
    let v = dotClumps2 norm clumpsX clumpsY params in
      ( addToCache xid yid v ; v )
