open Array
open Datapoint
open Util

(*let hhh = open_out "foo.foo"*)

let estimateBinary
  (d    : dataSet)
  (yval : float -> float option)  (* yval i should be +1 or -1 for the class of point i; or None if
                                     we shouldn't use it (for reasons *other* than CV -- we handle
                                     that internally); this is for multiclass.  so, say we're doing
                                     AVA and we're currently learning 5 versus 6.  yval 5 should be
                                     +1, yval 6 should be -1 and yval _ should be None *)
  (set  : int option)   (* for cross validation, train for all i!=set *)
  (eta  : float)        (* learning rate *)
  (numi : int)          (* number of iterations *)
  (eps  : float)        (* stop when |margin-1| <= eps *)
  (c    : float)        (* tradeoff *)
  : float * float array =
  Dot.clearCache ();
  if not d.dataHasClasses then failwith "trying to estimate params on data without classes" else
    let useThisOne i =
      match (set, d.crossValSplit) with
          (None, _) -> true
        | (Some v, Some (a,_)) -> a.(i) != v
        | _ -> failwith "trying to do cross validation without preparing for it" in

    let l = length d.datapoints in

    let bias  = ref 0.0 in
    let alpha = make l 0.0 in
    let used_points = make l 0 in
    let num_used = ref 1 in
    let all_yvals = make l 0. in
      for j = 0 to l - 1 do
	if useThisOne j then
          match yval d.dataclasses.(j) with
              Some yj -> ( 
		used_points.(j) <- 1 ;  
		all_yvals.(j) <- yj 
	      )
            | _ -> ();
      done;

    let iter = ref 0 in
    let margin = ref 0. in
    let curchar = ref 'a' in
    let next_char () = 
      match !curchar with
          'Z' -> 'Z'
        | 'z' -> curchar := 'A' ; 'z'
        | c   -> curchar := (Char.chr (1+Char.code c)) ; c in

    (* at first, all the zis will be zero.  then, after the zeroth iteration,
       i* will be (arbitrarily) 0 and so alpha_0 will become eta.  at this point
       we will need to update all the zs. *)

    let z = make l 0. in
    let istar = ref (Random.int l) in

      while (used_points.(!istar) == 0) do istar := Random.int l; done;

      (* begin the iterations *)
      while ((!iter < numi) && (fabs (!margin -. 1.) > eps)) do
        iter := !iter + 1;
	let _ = 
	  if !iter * 52 / numi != (!iter-1) * 52 / numi then (print_char (next_char ()); flush stdout) in
        (* update alpha.(k == !istar) *)
        let k = !istar in
(*        let _ = Printf.fprintf hhh "%d\n" k in  *)
        let yk = all_yvals.(k) in
        let ak = alpha.(k) in
        let oldBias = !bias in
        let deltaA = eta *. (1. -. yk *. z.(k)) in
        let factor =
          if alpha.(k) +. deltaA < 1e-100
          then ( bias := !bias -. yk *. ak        ; alpha.(k) <- 0.           ; 0. -. ak *. yk )
          else if alpha.(k) +. deltaA > c
          then ( bias := !bias +. yk *. (c -. ak) ; alpha.(k) <- c            ; (c -. alpha.(k)) *. yk )
          else ( bias := !bias +. yk *. ak        ; alpha.(k) <- ak +. deltaA ; deltaA *. yk ) in

        let deltaBias = !bias -. oldBias in

        (* update the zs *)
        let xk = d.datapoints.(k) in
        let minZplus  = ref infinity in
        let maxZminus = ref (0.-.infinity) in
        let minVal  = ref infinity in
        for i=0 to l-1 do
          if used_points.(i) > 0 then ( 
            let yi = all_yvals.(i) in
            z.(i) <- z.(i) +. factor *. Dot.dotClumps d.datanormalize xk d.datapoints.(i) k i d.trainingParams +. deltaBias;
            if yi *. z.(i) <= !minVal then ( minVal := yi *. z.(i) ; istar := i );
            if yi > 0. && z.(i) < !minZplus  then minZplus  := z.(i);
            if yi < 0. && z.(i) > !maxZminus then maxZminus := z.(i);
          );
        done;

        used_points.(k) <- used_points.(k) + 1;
        num_used := max !num_used used_points.(k);

        margin := 0.5 *. (!minZplus -. !maxZminus);
      done;
      
      for i=0 to l-1 do
	if nonZero alpha.(i) then
	  (match yval d.dataclasses.(i) with
	       None -> ()
	     | Some yi -> alpha.(i) <- alpha.(i) *. yi);
      done;
      
      print_char '\n'; flush stdout ;

      (!bias, alpha)

let estimateModel
  (str  : string)
  (d    : dataSet)
  (set  : int option)   (* for cross validation, train for all i!=set *)
  (eta  : float)        (* learning rate *)
  (numi : int)          (* number of iterations *)
  (eps  : float)        (* stop when |margin-1| <= eps *)
  (c    : float)        (* tradeoff *)
  (mc   : multiType option)
  : modelData =
  match mc with
      None ->
        let _ = Printf.printf "%s" str in
        let (bias, mults) = estimateBinary d (function i -> Some i) set eta numi eps c in
        let numSvs = 
          let tmp = ref 0 in
            for i=0 to length mults-1 do
              if nonZero mults.(i) then tmp := !tmp + 1;
            done;
            !tmp in
        let svs = init numSvs (function i -> [||]) in
        let m   = make numSvs 0. in
        let j   = ref 0 in
          for i=0 to length mults-1 do
            if nonZero mults.(i) then begin
              m.(!j)   <- mults.(i);
              svs.(!j) <- d.datapoints.(i);
              j := !j+1;
            end;
          done;
          { modelParams = d.trainingParams ;
            supportVectors = svs ;
            multipliers = [| m |] ;
            modelBias = [| bias |] ;
            modelnormalize = d.datanormalize;
            multiclass = None ;
            binarySigmoid = None ;
            multiSigmoid = None
          }

    | Some (OVA carr)
    | Some (OVN carr) ->
        let numC  = length carr in
        let isOVA = match mc with Some (OVA _) -> true | Some (OVN _) -> false | _ -> internal_error "mc not set" in
        let used  = make (length d.datapoints) false in
        let mults = init numC (function i -> [||]) in
        let bias  = make numC 0. in
        let getY cl i = 
          if int_of_float i == carr.(cl) then Some 1. 
          else if isOVA then Some (-1.)
          else None in

          for cl=0 to numC-1 do
            Printf.printf "%s  class %d: " str carr.(cl);
            let (b,m) = estimateBinary d (getY cl) set eta numi eps c in
              for j=0 to length m-1 do
                if nonZero m.(j) then used.(j) <- true;
              done;
              bias.(cl)  <- b;
              mults.(cl) <- m;
          done;
        let numSvs = 
          let tmp = ref 0 in
            for i=0 to length used-1 do
              if used.(i) then tmp := !tmp + 1;
            done;
            !tmp in
        let svs = init numSvs (function i -> [||]) in
        let m   = init numC   (function i -> make numSvs 0.) in
        let j   = ref 0 in
          for i=0 to length d.datapoints-1 do
            if used.(i) then begin
              for cl = 0 to numC-1 do
                m.(cl).(!j) <- mults.(cl).(i);
              done;
              svs.(!j)   <- d.datapoints.(i);
              j := !j+1;
            end;
          done;
          { modelParams = d.trainingParams ;
            supportVectors = svs ;
            multipliers = m ;
            modelBias = bias ;
            multiclass = mc ;
            modelnormalize = d.datanormalize;
            binarySigmoid = None ;
            multiSigmoid = None
          }

    | Some (AVA carr) ->
        let numC  = length carr in
        let used  = make (length d.datapoints) false in
        let mults = init (numC*(numC-1)/2) (function i -> [||]) in
        let bias  = make (numC*(numC-1)/2) 0. in
        let getY cl1 cl2 i = 
          if int_of_float i == carr.(cl1) then Some 1. 
          else if int_of_float i == carr.(cl2) then Some (-1.) 
          else None in

          for cl1=1 to numC-1 do
            for cl2=0 to cl1-1 do
              Printf.printf "%s  class %d/%d (%d): " str (min carr.(cl1) carr.(cl2)) (max carr.(cl1) carr.(cl2)) (avaCLposition cl1 cl2);
              let (b,m) = estimateBinary d (getY cl1 cl2) set eta numi eps c in
                for j=0 to length m-1 do
                  if nonZero m.(j) then used.(j) <- true;
                done;
                bias.(avaCLposition cl1 cl2)  <- b;
                mults.(avaCLposition cl1 cl2) <- m;
            done;
          done;
        let numSvs = 
          let tmp = ref 0 in
            for i=0 to length used-1 do
              if used.(i) then tmp := !tmp + 1;
            done;
            !tmp in
        let svs = init numSvs (function i -> [||]) in
        let m   = init (numC*(numC-1)/2) (function i -> make numSvs 0.) in
        let j   = ref 0 in
          for i=0 to length d.datapoints-1 do
            if used.(i) then begin
              for cl1 = 1 to numC-1 do
                for cl2=0 to cl1-1 do
                  m.(avaCLposition cl1 cl2).(!j) <- mults.(avaCLposition cl1 cl2).(i);
                done;
              done;
              svs.(!j)   <- d.datapoints.(i);
              j := !j+1;
            end;
          done;
          { modelParams = d.trainingParams ;
            supportVectors = svs ;
            multipliers = m ;
            modelnormalize = d.datanormalize;
            modelBias = bias ;
            multiclass = mc ;
            binarySigmoid = None ;
            multiSigmoid = None
          }

let crossValidate
  (d    : dataSet)
  (eta  : float)        (* learning rate *)
  (numi : int)          (* number of iterations *)
  (eps  : float)        (* stop when |margin-1| <= eps *)
  (c    : float)        (* tradeoff *)
  (mc   : multiType option)
  : int =
  match d.crossValSplit with
      None -> failwith "cannot cross validate on data which has not been prepared"
    | Some (cv,nfolds) ->
        let err = ref 0 in
          for i = 0 to nfolds - 1 do
            let m = estimateModel ("Fold " ^ string_of_int i ^ ": ") d (Some i) eta numi eps c mc in
            err := !err + Predict.predictCV m d i;
          done;
          !err

let testDataSet =
  {
    trainingParams = [| DVP (1.0, LinearKernel) |];
    datapoints = [| [| DV [|  1.0 ;  2.0 |] |] ;
		    [| DV [|  1.2 ;  2.1 |] |] ;
		    [| DV [| -1.0 ; -2.1 |] |] ;
		    [| DV [| -1.2 ; -2.0 |] |] |];
    dataclasses = [| 1. ; 1. ; -1. ; -1. |];
    datanormalize = None;
    rankingPoints = None;
    dataHasClasses = true;
    crossValSplit = None;
  }
          
(*                  
let m = estimateModel testDataSet None 0.1 1000 0.0001 1. None
let _ = BinaryIO.saveModel "test.model" m
*)
(*
let m = BinaryIO.loadModel "test.model"
let _ =
  Printf.printf "%d\n" (snd (Predict.predictOne m testDataSet.datapoints.(0)));
  Printf.printf "%d\n" (snd (Predict.predictOne m testDataSet.datapoints.(1)));
  Printf.printf "%d\n" (snd (Predict.predictOne m testDataSet.datapoints.(2)));
  Printf.printf "%d\n" (snd (Predict.predictOne m testDataSet.datapoints.(3)))
*)
