open Datapoint
open Array
open Util

(* so far, we only support normalization of columns; i.e., mapping
   each (real) feature in the training set down so that it falls in
   [-1,1]; we do this by finding the most negative value and the most
   positive value, and dividing everything by the max of the absolute
   values of those *)

let normalizeDataBy (d : dataSet) (norm : float array array) : unit =
  let numPts    = length d.datapoints in
  let numClumps = length d.trainingParams in
  let numF      = make numClumps 0 in
    for n = 0 to numPts-1 do
      for c = 0 to numClumps-1 do
        match d.datapoints.(n).(c) with
            DV v -> if length norm.(c) == 0 then failwith "invalid normalization combination" else
              ( for i = 0 to length v - 1 do
                  if norm.(c).(i) > 0. then v.(i) <- v.(i) /. norm.(c).(i);
                done )
          | SV (f,v) -> if length norm.(c) == 0 then failwith "invalid normalization combination" else
              ( let hitNeg = ref false in
                  for i=0 to length f-1 do
                    if f.(i) < 0 then hitNeg := true;
                    if not !hitNeg && norm.(c).(f.(i)) > 0. then v.(i) <- v.(i) /. norm.(c).(f.(i));
                  done )
          | _ -> ();
      done;
    done;
    d.datanormalize <- Some norm;
    ()

let normalizeDataBySelf (d : dataSet) : unit =
  if isSome d.datanormalize then failwith "data is already normalized" else
    let numPts    = length d.datapoints in
    let numClumps = length d.trainingParams in
    let numF      = make numClumps 0 in
      
      (* find the max feature number *)
      for n = 0 to numPts-1 do
        for c = 0 to numClumps-1 do
          match d.datapoints.(n).(c) with
              DV v -> numF.(c) <- max numF.(c) (length v)
            | SV (f,_) ->
                ( let hitNeg = ref false in
                    for i=0 to length f-1 do
                      if f.(i) < 0 then hitNeg := true;
                      if not !hitNeg then numF.(c) <- max numF.(c) f.(i);
                    done )
            | DP _ -> numF.(c) <- 1;
            | ST _ -> numF.(c) <- 1;
            | _ -> ();
        done;
      done;

    let norm = init numClumps (fun i -> make (1+numF.(i)) 0.) in

      (* find the (abs) max value for each feature *)
      for n = 0 to numPts-1 do
        for c = 0 to numClumps-1 do
          match d.datapoints.(n).(c) with
              DV v -> 
                ( for i = 0 to length v - 1 do
                    norm.(c).(i) <- max norm.(c).(i) (fabs v.(i));
                  done )
            | SV (f,v) ->
                ( let hitNeg = ref false in
                    for i=0 to length f-1 do
                      if f.(i) < 0 then hitNeg := true;
                      if not !hitNeg then norm.(c).(f.(i)) <- max norm.(c).(f.(i)) (fabs v.(i));
                    done )
            | DP str ->
                let (lambda,maxn) = match d.trainingParams.(c) with DPP (_,l,n) -> (l,n) | _ -> internal_error "clump and data mismatch" in
                  norm.(c).(0) <- Dot.dpStringKernel lambda maxn str str
            | ST str ->
                let wt = match d.trainingParams.(c) with STP (_,w) -> w | _ -> internal_error "clump and data mismatch" in
                  norm.(c).(0) <- Dot.stStringKernel wt str str
            | _ -> ()
        done;
      done;

      (* normalize *)
      normalizeDataBy d norm; (* this also sets the datanormalize value *)
      ()

