open Datapoint
open Array
open Str
open Util

module H = Hashtbl

let version = "0.0.1"

(* this is the module that tells us how to run scripts *)

type datatype = SVMData | SepData of regexp

(* a script is just a sequence (list) of commands *)

let (dataEnv  : (string, dataSet)   H.t) = H.create 2
let (modelEnv : (string, modelData) H.t) = H.create 2
let (valEnv   : (string, string)    H.t) = H.create 10

let getValEnv str =
  if H.mem valEnv str then H.find valEnv str
  else 
    try
      Sys.getenv str
    with
        Not_found -> failwith ("variable $" ^ str ^ " not defined")
      | e -> raise e

let (quiet, versiononly, usage, inputfile) =
  let q   = ref false in
  let vo  = ref false in
  let us  = ref false in
  let inp = ref None  in
  let i   = ref 1 in
    while (!i < length Sys.argv) && (String.get Sys.argv.(!i) 0 == '-') do
      (match Sys.argv.(!i) with
           "-q" -> q := true
         | "-v" -> vo := true
         | "-h"
         | "-?"
         | "--help" -> us := true
         | x    -> failwith ("unknown command line argument '" ^ x ^ "', try '-h' for help"));
      i := !i + 1;
    done;
    if (!i < length Sys.argv) then begin
      inp := Some (Sys.argv.(!i));

      let posn = ref 1 in
      for j=1 to length Sys.argv-1 - !i do
        if (String.get Sys.argv.(j + !i) 0 == '@' && String.contains Sys.argv.(j + !i) '=')
        then (
          let idx = String.index Sys.argv.(j + !i) '=' in
          let nam = String.sub   Sys.argv.(j + !i) 1 (idx-1) in
          let va  = String.sub   Sys.argv.(j + !i) (idx+1) (String.length Sys.argv.(j + !i) - idx - 1) in
            H.replace valEnv nam va;
        )
        else (
          H.replace valEnv (string_of_int !posn) Sys.argv.(j + !i);
          posn := !posn+1;
        );
      done;
    end;
    (!q, !vo, !us, !inp)

let printusage () =
  Printf.printf "Usage: %s [-q] [-v] [-h] [inputfile] [arguments to inputfile]\n" Sys.argv.(0);
(*  Printf.printf "   or: %s [-q] -c <command>\n\n" Sys.argv.(0); *)
  Printf.printf "  -q : run (moderately) silently\n";
  Printf.printf "  -v : print version and quit\n";
(*  Printf.printf "  -c : run a single command (or multiple, separated with ';')\n" *)
  Printf.printf "  -h : show this usage information\n\n"

let shouldOutput = ref (not quiet)
let silence = shouldOutput := false
let unsilence = shouldOutput := not quiet

(*let dprintf fmt =
 if !shouldOutput then Printf.kprintf (fun s -> print_string s ; flush stdout) fmt else Printf.kprintf ignore fmt;;
let dprintf = Printf.printf *)

let deleteData name = H.remove dataEnv name

let orelse f x v = match x with None -> v | Some x -> f x

let rec findArg str = function
    [] -> None
  | [x] -> if x = str then failwith ("option '" ^ str ^ "' requires an argument") else None
  | (x::y::xs) -> if x = str then Some y else findArg str (y::xs)

let ensureData name =
  if H.mem dataEnv name then () else failwith ("data '" ^ name ^ "' does not exist")

let ensureModel name =
  if H.mem modelEnv name then () else failwith ("model '" ^ name ^ "' does not exist")

let kernel_info = function
    LinearKernel -> "linear"
  | PolynomialKernel (a,b,d) -> "poly constant=" ^ string_of_float a ^ 
                                "coefficient=" ^ string_of_float b ^
                                "degree=" ^ string_of_float d
  | RbfKernel g -> "rbf gamma=" ^ string_of_float g
  | SigKernel (a,b) -> "sig constant=" ^ string_of_float a ^ 
                       "coefficient=" ^ string_of_float b
  | IdKernel b -> "id beta=" ^ string_of_float b

let showDataInfo name =
  ensureData name;
  let ds = H.find dataEnv name in
    dprintf "%s has %d clumps and %d points:\n" name (length ds.trainingParams) (length ds.datapoints);
    for i=0 to length ds.trainingParams-1 do
      match ds.trainingParams.(i) with
          DVP (w,k)    -> dprintf "  %d: dense   , weight=%g, kernel=%s\n" i w (kernel_info k)
        | SVP (w,k,h)  -> dprintf "  %d: sparse  , weight=%g, kernel=%s, highest=%d\n" i w (kernel_info k) h
        | DPP (w,l,n)  -> dprintf "  %d: dpstring, weight=%g, lambda=%g, maxn=%d\n" i w l n
        | STP (w,f)    -> dprintf "  %d: ststring, weight=%g, %s\n" i w (match f with SuffixTreeKernel.ExpDecay l -> "expdecay " ^ string_of_float l | SuffixTreeKernel.ConstWeight -> "const" | SuffixTreeKernel.BoundedRange n -> "bounded " ^ string_of_int n)
        | DiscP (w,a,b)-> dprintf "  %d: discrete, weight=%g, cardinality=%d, beta=%g\n" i w a b
        | MatP (w,_)   -> dprintf "  %d: matrix,   weight=%g\n" i w
        | FVParamPlaceHolder -> dprintf "  %d: unknown\n" i;
    done

let newData name =
  if H.mem dataEnv name 
  then failwith ("cannot create data '" ^ name ^ "'; name already in use, delete it first")
  else 
    dprintf "Data %s added\n" name;
    H.add dataEnv name
      { trainingParams = [||] ;
        datanormalize = None;
        datapoints = [||] ;
        dataclasses = [||] ;
        rankingPoints = None ;
        crossValSplit = None ;
        dataHasClasses = false }

let modClump name num args =
  ensureData name ;
  let ds = H.find dataEnv name in
  let l  = length ds.trainingParams in
  if num >= l then failwith ("clump " ^ string_of_int num ^ " does not exist in '" ^name^"'") else
    let cl =
      match ds.trainingParams.(num) with
          FVParamPlaceHolder -> FVParamPlaceHolder
        | DVP (w,k) -> DVP (orelse float_of_string (findArg "weight" args) w,
                            match findArg "kernel" args with
                                None -> k
                              | Some "linear" -> LinearKernel
                              | Some "poly"   -> PolynomialKernel (orelse float_of_string (findArg "constant" args) 0.,
                                                                   orelse float_of_string (findArg "coefficient" args) 1.,
                                                                   orelse float_of_string (findArg "degree" args) 2.)
                              | Some "rbf"    -> RbfKernel (orelse float_of_string (findArg "gamma" args) 1.)
                              | Some "sig"    -> SigKernel (orelse float_of_string (findArg "constant" args) 0.,
                                                            orelse float_of_string (findArg "coefficient" args) 1.)
                              | Some "id"     -> IdKernel  (orelse float_of_string (findArg "beta" args) 0.5)
                              | Some k -> failwith ("unknown kernel type: " ^ k))
        | SVP(w,k,h) -> SVP (orelse float_of_string (findArg "weight" args) w,
                             (match findArg "kernel" args with
                                  None -> k
                                | Some "linear" -> LinearKernel
                                | Some "poly"   -> PolynomialKernel (orelse float_of_string (findArg "constant" args) 0.,
                                                                     orelse float_of_string (findArg "coefficient" args) 1.,
                                                                     orelse float_of_string (findArg "degree" args) 2.)
                                | Some "rbf"    -> RbfKernel (orelse float_of_string (findArg "gamma" args) 1.)
                                | Some "sig"    -> SigKernel (orelse float_of_string (findArg "constant" args) 0.,
                                                              orelse float_of_string (findArg "coefficient" args) 1.)
                                | Some "id"     -> IdKernel  (orelse float_of_string (findArg "beta" args) 0.5)
                                | Some k -> failwith ("unknown kernel type: " ^ k)),
                             orelse int_of_string (findArg "highest" args) h)
        | DPP (w,l,n) -> DPP (orelse float_of_string (findArg "weight" args) w,
                              orelse float_of_string (findArg "lambda" args) l,
                              orelse int_of_string (findArg "maxn" args) n)
        | STP (w,f) -> STP (orelse float_of_string (findArg "weight" args) w,
			    match (findArg "expdecay" args, List.mem "const" args, findArg "bounded" args) with
				(Some l, false, None  ) -> SuffixTreeKernel.ExpDecay (float_of_string l)
			      | (None,   true,  None  ) -> SuffixTreeKernel.ConstWeight
			      | (None,   false, Some n) -> SuffixTreeKernel.BoundedRange (int_of_string n)
			      | (None,   false, None  ) -> SuffixTreeKernel.ExpDecay 0.75
			      | _ -> failwith "invalid combination of suffix tree types")
        | DiscP (w,c,l) -> DiscP (orelse float_of_string (findArg "weight" args) w,
                                  orelse int_of_string (findArg "cardinality" args) c,
                                  orelse float_of_string (findArg "lambda" args) l)
        | MatP (w,m)   -> MatP (orelse float_of_string (findArg "weight" args) w, m) in
      ds.trainingParams.(num) <- cl;
      dprintf "Parameters of clump #%d of %s set\n" num name

let addClump name clumpType args =
  ensureData name ;
    let ds = H.find dataEnv name in
      if length ds.datapoints > 0 then failwith "all clumps must be added before data is loaded" else
        let cl = 
          match clumpType with
              "dense"    -> DVP   (1., LinearKernel)
            | "sparse"   -> SVP   (1., LinearKernel, -1)
            | "dpstring" -> DPP   (1., 0.75, 4)
            | "ststring" -> STP   (1., SuffixTreeKernel.ExpDecay 0.75)
            | "discrete" -> DiscP (1., -1, 0.5)
            | "matrix"   -> MatP  (1., [|[||]|])
            | _          -> failwith ("unknown clump type '" ^ clumpType ^ ", must be one of dense, sparse, dpstring, ststring, discrete or matrix") in
        let a = make (1+length ds.trainingParams) FVParamPlaceHolder in
          for i=0 to length ds.trainingParams-1 do
            a.(i) <- ds.trainingParams.(i);
          done;
          a.(length ds.trainingParams) <- cl;
          dprintf "Clump #%d added to %s\n" (length ds.trainingParams) name;
          H.replace dataEnv name ({ ds with trainingParams = a });
          silence;
          modClump name (length ds.trainingParams) args;
          unsilence

let scanFile filepath typ =
  let (hasC, lc, lowF, highF) =
    match typ with
        "svm" -> DataIO.getFileInfoSVM filepath
      | "csv" -> DataIO.getFileInfoSeparated (regexp ",") filepath
      | "ssv" -> DataIO.getFileInfoSeparated (regexp "[ \t]+") filepath
      | _     -> failwith "unknown filetype, must be one of svm, csv, ssv" in
    dprintf "File '%s' %s, has %d points, lowest feature is %d, highest is %d\n"
      filepath (if hasC then "has classes" else "doesn't have classes") lc lowF highF

let readColumn (lf,hf) x = 
  let rec parse_columns = function
      [] -> []
    | (Text n :: Delim "-" :: Text m :: []) -> (int_of_string n, int_of_string m) :: []
    | (Text n :: Delim "-" :: Text m :: Delim "," :: xs) -> (int_of_string n, int_of_string m) :: parse_columns xs
    | (Text n :: []) -> (int_of_string n, int_of_string n) :: []
    | (Text n :: Delim "," :: xs) -> (int_of_string n, int_of_string n) :: parse_columns xs
    | _ -> failwith ("malformed column info: '" ^ x ^ "'") in
    if x = "all" then [(lf,hf)]
    else
      List.sort (function (i,_) -> function (j,_) -> compare i j) 
        (parse_columns (full_split (regexp "[-,]") x))

let rec getColInfo (lf,hf) = function
    [] -> []
  | ("clump"::clump::"from"::cols::xs) -> (int_of_string clump, readColumn (lf,hf) cols) :: getColInfo (lf,hf) xs
  | ("clump"::_) -> failwith "cannot parse column/clump information"
  | (_::xs) -> getColInfo (lf,hf) xs

let rec getLength acc = function
    [] -> acc
  | ((l,h)::xs) -> getLength (acc+h-l+1) xs

let rec getFNum f acc = function
    [] -> None
  | ((l,h)::xs) ->
      if l <= f && f <= h
      then Some (acc+f-l, getLength (acc+h-l+1) xs)
      else getFNum f (acc+h-l+1) xs

(* first component is clump, second is feature number there, third is total length *)
let rec findColumn f = function
    [] -> None
  | ((i, l)::xs) ->
      match getFNum f 0 l with
          None -> findColumn f xs
        | Some (j,l) -> Some (i,j,l)

let splitParens str = global_replace (regexp "(") "#(#" (global_replace (regexp ")") "#)#" str)

let makeST wt str =
  let t   = SuffixTree.createTree str in
(*  let _   = SuffixTree.printTree t in *)
  let kvc = SuffixTreeKernel.computeKernelValCache (SuffixTreeKernel.weightTypeToFn wt) t in
  let str_with_neg1 = init (length str+1) (fun i -> if i < length str then str.(i) else -1) in
    { SuffixTreeKernel.tree_ref    = t ;
      SuffixTreeKernel.tree_kvc    = kvc }

let loadFile name filepath typ columns = 
  ensureData name;
  let ds = H.find dataEnv name in
  let _  = dprintf "Loading data from %s into '%s'\n" filepath name in
  let curRow  = ref (-1) in
  let genericAddFV columns i f v =
    match findColumn f columns with
        None -> ()
      | Some (clump, fn, totalLen) ->
          if clump >= length ds.trainingParams then failwith "attempting to load data into undefined clump" else
          match ds.trainingParams.(clump) with
              FVParamPlaceHolder -> ()
            | DVP _ -> 
                if !curRow != i then ( ds.datapoints.(i).(clump) <- DV (make totalLen 0.) ; curRow := i );
                let a = match ds.datapoints.(i).(clump) with DV a -> a | _ -> failwith "invalid clump type" in 
                  a.(fn) <- float_of_string v
            | SVP (w,k,hfn) ->
                (* if we're creating a new row, make the arrays *)
                if !curRow != i 
                then ( let a = make 32 (-1) in
                         a.(31) <- 0;  (* the last component (for now) stores the current location;
                                          yes, it's a huge hack, but... *)
                         ds.datapoints.(i).(clump) <- SV (a, make 32 0.) ; 
                         curRow := i );
                if nonZero (float_of_string v) then begin
                  let (p,fa,va) = 
                    let (oldF,oldV) = match ds.datapoints.(i).(clump) with SV a -> a | _ -> failwith "invalid data type" in
                    let len = length oldF in

                    (* expand if we need to *)
                    let p = oldF.(len-1) in
                      if p >= length oldF-2
                      then 
                        let newF = make ((length oldF + 5) * 2) (-1)  in
                        let newV = make ((length oldF + 5) * 2) 0. in
                          for i=0 to length oldF-1 do
                            newF.(i) <- oldF.(i) ; newV.(i) <- oldV.(i);
                          done;
                          newF.(len-1) <- -1;
                          ds.datapoints.(i).(clump) <- SV (newF, newV);
                          (p, newF, newV)
                      else (p, oldF, oldV) in

                    (* set the value *)
                    fa.(p) <- fn;
                    va.(p) <- float_of_string v;

                    (* update the position *)
                    fa.(length fa-1) <- p+1;

                    (* update the highest feature number *)
                    if fn > hfn then ds.trainingParams.(clump) <- SVP (w,k,fn);
                end;

            | DPP _ -> ds.datapoints.(i).(clump) <- DP (of_list (List.map (Vocabulary.get Vocabulary.global) (split (regexp "#") (splitParens v))))
            | STP (_,wt) -> ds.datapoints.(i).(clump) <- ST (makeST wt (of_list (List.map (Vocabulary.get Vocabulary.global) (split (regexp "#") (splitParens v)))))
            | DiscP _ -> ds.datapoints.(i).(clump) <- Disc (Vocabulary.get Vocabulary.global v)
            | MatP _ -> failwith "trying to assign data to matrix; use 'load matrix' to load a matrix" in

  match typ with
      "svm" -> begin
        let hascl   = not (List.mem "noclasses" columns) in
        let ranked  = List.mem "rank" columns in
        let lowF    = findArg "low" columns in
        let highF   = findArg "high" columns in
        let numEx   = findArg "points" columns in
        let (h,lc,lf,hf) =
          if isNone lowF || isNone highF || isNone numEx
          then DataIO.getFileInfoSVM filepath
          else (hascl, int_of_string (fromSome numEx), int_of_string (fromSome lowF), int_of_string (fromSome highF)) in
        let rankValues = if ranked then make lc (-1) else [| |] in
        let setClass i f = if i >= length ds.dataclasses then failwith "trying to set class of undefined datapoint" else ds.dataclasses.(i) <- f in
        let setRank i j  = if i >= length rankValues then failwith "trying to set rank of undefined datapoint" else rankValues.(i) <- j in
        let columns = getColInfo (lf,hf) columns in
	let _ = if h then ds.dataHasClasses <- true in
        let addFV i f v = genericAddFV columns i f v in

          if h then ds.dataclasses <- make lc 0.;
          ds.datapoints  <- init lc (function _ -> make (length ds.trainingParams) (Mat (-1)));
          
          DataIO.addToDataSVM filepath h ranked setClass setRank addFV;
          dprintf "Data loaded from %s into '%s'\n" filepath name
      end

    | "csv" 
    | "ssv" -> begin
        let re = if typ = "csv" then regexp "," else regexp "[ \t]+" in
        let classColumn = findArg "classes" columns in
        let lowF    = findArg "low" columns in
        let highF   = findArg "high" columns in
        let numEx   = findArg "points" columns in

        let isClassColumn c =
          match classColumn with
              None -> false
            | Some s -> c == int_of_string s in

        let (h,lc,lf,hf) =
          if isNone lowF || isNone highF || isNone numEx
          then DataIO.getFileInfoSeparated re filepath
          else (isSome classColumn, int_of_string (fromSome numEx), int_of_string (fromSome lowF), int_of_string (fromSome highF)) in

        let columns = getColInfo (lf,hf) columns in
	let _ = if h then ds.dataHasClasses <- true in

        let addFV i f v =
          if isClassColumn f
          then if i >= length ds.dataclasses then failwith "trying to set class of undefined datapoint" else ds.dataclasses.(i) <- float_of_string v
          else genericAddFV columns i f v in

          if h then ds.dataclasses <- make lc 0. ;
          ds.datapoints <- init lc (function _ -> make (length ds.trainingParams) (Mat (-1)));

          DataIO.addToDataSeparated re filepath addFV;
          dprintf "Data loaded from %s into '%s'\n" filepath name
      end

    | _ -> failwith ("unknown file type '" ^ typ ^ "', must be one of svm, csv, ssv")

let countClasses ds =
  let h = H.create 4 in
    for i=0 to length ds-1 do
      H.replace h (int_of_float ds.(i)) ();
    done;
  let sz = hash_size h in 
  let a  = make sz 0 in
  let _ = H.fold (fun i _ c -> a.(c) <- i ; c+1) h 0 in
    sort compare a;
    a

let setupXV name n =
  ensureData name;
  let ds = H.find dataEnv name in
  if n <= 1 then 
    ( ds.crossValSplit <- None;
      dprintf "Cross validation turned off for '%s'\n" name )
  else
    ( let arr = make (length ds.dataclasses) 0 in
	for i=0 to length ds.dataclasses-1 do
	  arr.(i) <- Random.int n;
	done;
	ds.crossValSplit <- Some (arr, n);
	dprintf "Cross validation for %d folds initialized for '%s'\n" n name )
    
let estModel modelname name args =
  ensureData name ;
  let d    = H.find dataEnv name in
  let eta  = orelse float_of_string  (findArg "eta" args)  0.1 in
  let numi = orelse int_of_string    (findArg "numi" args) 1000 in
  let eps  = orelse float_of_string  (findArg "epsilon" args) 0.0001 in
  let c    = orelse float_of_string  (findArg "c" args) 1. in
  let mc   = 
    if List.mem "ova" args then Some (OVA (countClasses d.dataclasses))
    else if List.mem "ava" args then Some (AVA (countClasses d.dataclasses))
    else if List.mem "ovn" args then Some (OVN (countClasses d.dataclasses))
    else if length (countClasses d.dataclasses) > 2 then failwith "need to specify multiclass type (ova,ovn,ava) for mc data"
    else None in
  let model = Estimate.estimateModel "" d None eta numi eps c mc in
    dprintf "Model %s estimated from %s\n" modelname name;
    H.replace modelEnv modelname model

let showModelInfo mname = 
  ensureModel mname ;
  let m = H.find modelEnv mname in
    dprintf "%s has %d clumps:\n" mname (length m.modelParams);
    for i=0 to length m.modelParams-1 do
      match m.modelParams.(i) with
          DVP (w,k)    -> dprintf "  %d: dense   , weight=%g, kernel=%s\n" i w (kernel_info k)
        | SVP (w,k,h)  -> dprintf "  %d: sparse  , weight=%g, kernel=%s, highest=%d\n" i w (kernel_info k) h
        | DPP (w,l,n)  -> dprintf "  %d: dpstring, weight=%g, lambda=%g, maxn=%d\n" i w l n
        | STP (w,f)    -> dprintf "  %d: ststring, weight=%g, %s\n" i w (match f with SuffixTreeKernel.ExpDecay l -> "expdecay " ^ string_of_float l | SuffixTreeKernel.ConstWeight -> "const" | SuffixTreeKernel.BoundedRange n -> "bounded " ^ string_of_int n)
        | DiscP (w,a,b)-> dprintf "  %d: discrete, weight=%g, cardinality=%d, beta=%g\n" i w a b
        | MatP (w,_)   -> dprintf "  %d: matrix,   weight=%g\n" i w
        | FVParamPlaceHolder -> dprintf "  %d: unknown\n" i;
    done;
    dprintf "%d support vectors, %s %s\n" 
      (length m.supportVectors) 
      (match m.multiclass with
	   None -> "binary classifier"
	 | Some (OVA n) -> "one-versus-all (" ^ string_of_int (length n) ^ ") classifier"
	 | Some (AVA n) -> "all-versus-all (" ^ string_of_int (length n) ^ ") classifier"
	 | Some (OVN n) -> "one-versus-none (" ^ string_of_int (length n) ^ ") classifier")
      (if not (isNone m.binarySigmoid) then ", probabilistic"
       else if not (isNone m.multiSigmoid) then ", probabilistic"
       else "")

let pred name mname args =
  ensureData  name;
  ensureModel mname;
  let outf = findArg "output" args in
  let outh = 
    match outf with
	None -> None
      | Some "-" -> Some stdout
      | Some fn  -> Some (open_out fn) in
  let d = H.find dataEnv  name  in
  let m = H.find modelEnv mname in 
  let err = ref 0 in
  let n = length d.datapoints in

    for i=0 to n - 1 do
      let truth = d.dataclasses.(i) in
      let (arr,pred) = Predict.predictOne m d.datapoints.(i) d.datanormalize in
      let trueClass i =
	match m.multiclass with
	    None -> string_of_int i
	  | Some (OVA a) -> string_of_int a.(i)
	  | Some (OVN a) -> string_of_int a.(i)
	  | Some (AVA a) -> (
	      let (x,y) = unAVACLposition i in 
		string_of_int a.(x) ^ ":" ^ string_of_int a.(y)) in

	if int_of_float truth != pred then err := !err+1;
	(match outh with
	     Some h -> 
	       Printf.fprintf h "%d" pred;
               if isNone m.multiclass
               then begin
                 Printf.fprintf h " %g" arr.(0);
                 (match m.binarySigmoid with
                      None -> ()
                    | Some (a,b) -> Printf.fprintf h " p=%g" (1. /. (1. +. exp (a *. arr.(0) +. b))));
               end else begin
	         for i=0 to length arr-1 do
		   Printf.fprintf h " %s:%g" (trueClass i) arr.(i);
	         done
               end;
	       Printf.fprintf h "\n";
	   | _ -> ());
    done;
    dprintf "Accuracy = %d/%d = %g%c\n" (n - !err) n (100. *. float_of_int (n - !err) /. float_of_int n) '%';
    (match (outf,outh) with
       | (Some "-",_) -> ()
       | (Some _,Some h) -> close_out h
       | _ -> ())

let isDigit c = Char.code c >= 48 && Char.code c <= 57

let rec takeWhile f = function
    [] -> []
  | (x::xs) -> if f x then x :: takeWhile f xs else []

let rec drop n = function
    [] -> []
  | (x::xs) -> if n <= 0 then x::xs else drop (n-1) xs

let allArguments low = 
  let l = ref [] in
  let i = ref low  in
    while (H.mem valEnv (string_of_int !i)) do
      l := H.find valEnv (string_of_int !i) :: !l;
      i := !i + 1;
    done;
    String.concat " " (List.rev !l)


let implode l =
  let n = List.length l in
  let str = String.make n ' ' in
  let i = ref 0 in
    List.iter (fun c -> String.set str !i c ; i := !i+1) l;
    str
      
let rec last = function
    [] -> internal_error "last on []"
  | [x] -> x
  | (_::xs) -> last xs

let rec init_list = function
    [] -> []
  | [_] -> []
  | (x::xs) -> x :: init_list xs

let isAlphaNum c = 
  let o = Char.code c in
    (97 <= o && o <= 122) || (65 <= o && o <= 90) || (48 <= o && o <= 57)

let rec replaceVariables = function
    [] -> []
  | ('\\'::'$'::xs) -> "$" :: replaceVariables xs
  | ('$'::'*'::xs)  -> allArguments 1 :: replaceVariables xs
  | ('$'::'0'::xs)  -> Sys.argv.(0) :: replaceVariables xs
  | ('$'::'?'::'*'::xs) -> allArguments 1 :: replaceVariables xs
  | ('$'::'?'::'0'::xs)  -> Sys.argv.(0) :: replaceVariables xs
  | ['$']           -> failwith "expecting variable name after '$'"
  | ['$';'?']       -> failwith "expecting variable name after '$?'"
  | ('$'::'?'::xs)  ->
      let rest = ref [] in
      let v =
        try
          if isDigit (List.hd xs)
          then 
	    let n = takeWhile (fun c -> isDigit c || c == '+') xs in
	    let _ = rest := drop (List.length n) xs in
              if last n == '+'
              then allArguments (int_of_string (implode (init_list n)))
              else getValEnv (implode n)
          else
	    let s = takeWhile (fun c -> isAlphaNum c) xs in
	    let _ = rest := drop (List.length s) xs in
              getValEnv (implode s)
        with
            Failure _ ->  "" 
          | e -> raise e in
        v :: replaceVariables !rest
              
  | ('$'::xs)       ->
      if isDigit (List.hd xs)
      then 
	let n = takeWhile (fun c -> isDigit c || c == '+') xs in
	let rest = drop (List.length n) xs in
          if last n == '+'
          then
            allArguments (int_of_string (implode (init_list n))) :: replaceVariables rest
          else
	    getValEnv (implode n) :: replaceVariables rest
      else
	let s = takeWhile (fun c -> isAlphaNum c) xs in
	let rest = drop (List.length s) xs in
	  getValEnv (implode s) :: replaceVariables rest
  | (x::xs) -> (String.make 1 x) :: replaceVariables xs
      
let explode str =
  let l = ref [] in
  let n = String.length str in
    for i = 0 to n-1 do
      l := (String.get str (n-1-i)) :: !l;
    done;
    !l

let rec getOptParams ds = function
    [] -> []
  | (x::xs) ->
      begin
        let setP clumpId paramName v    = 
          ds.trainingParams.(clumpId) <-
            (match (ds.trainingParams.(clumpId), paramName) with
                 (FVParamPlaceHolder,_) -> FVParamPlaceHolder
                   
               | (DVP (w,k), "weight")  -> DVP (v,k)
               | (DVP (w,PolynomialKernel (a,b,d)), "constant"   ) -> DVP (w,PolynomialKernel (v,b,d))
               | (DVP (w,PolynomialKernel (a,b,d)), "coefficient") -> DVP (w,PolynomialKernel (a,v,d))
               | (DVP (w,PolynomialKernel (a,b,d)), "degree"     ) -> DVP (w,PolynomialKernel (a,b,v))
               | (DVP (w,RbfKernel        g      ), "gamma"      ) -> DVP (w,RbfKernel v)
               | (DVP (w,SigKernel        (a,b)  ), "constant"   ) -> DVP (w,SigKernel (v,b))
               | (DVP (w,SigKernel        (a,b)  ), "coefficient") -> DVP (w,SigKernel (a,v))
               | (DVP (w,IdKernel         b      ), "beta"       ) -> DVP (w,IdKernel v)

               | (SVP (w,k,h), "weight")  -> SVP (v,k,h)
               | (SVP (w,PolynomialKernel (a,b,d),h), "constant"   ) -> SVP (w,PolynomialKernel (v,b,d),h)
               | (SVP (w,PolynomialKernel (a,b,d),h), "coefficient") -> SVP (w,PolynomialKernel (a,v,d),h)
               | (SVP (w,PolynomialKernel (a,b,d),h), "degree"     ) -> SVP (w,PolynomialKernel (a,b,v),h)
               | (SVP (w,RbfKernel        g      ,h), "gamma"      ) -> SVP (w,RbfKernel v,h)
               | (SVP (w,SigKernel        (a,b)  ,h), "constant"   ) -> SVP (w,SigKernel (v,b),h)
               | (SVP (w,SigKernel        (a,b)  ,h), "coefficient") -> SVP (w,SigKernel (a,v),h)
               | (SVP (w,IdKernel         b      ,h), "beta"       ) -> SVP (w,IdKernel v,h)

               | (DPP (w,l,n), "weight") -> DPP (v,l,n)
               | (DPP (w,l,n), "lambda") -> DPP (w,v,n)
               | (DPP (w,l,n), "maxn")   -> DPP (w,l,int_of_float v)

               | (STP (w,f), "weight") -> STP (v,f)
	       | (STP (w,SuffixTreeKernel.ExpDecay _), "lambda") -> STP (w, SuffixTreeKernel.ExpDecay v)
	       | (STP (w,SuffixTreeKernel.BoundedRange _), "bounded") -> STP (w, SuffixTreeKernel.BoundedRange (int_of_float v))
	       | (STP _, s) -> failwith ("cannot optimize parameter '" ^ s ^ "' for suffix tree")

               | (DiscP (w,c,l), "weight"     ) -> DiscP (v,c,l)
               | (DiscP (w,c,l), "cardinality") -> DiscP (w,int_of_float v,l)
               | (DiscP (w,c,l), "lambda"     ) -> DiscP (w,c,v)

               | (MatP (w,m), "weight") -> MatP (v,m)

               | (_, arg) -> failwith ("optimize: '" ^ arg ^ "' is not a valid argument for clump " ^ string_of_int clumpId)) in

          if Str.string_match (regexp ("^\\([0-9]+\\):\\([^:]+\\):log2:\\(-?[0-9\\.]+\\):\\(-?[0-9\\.]+\\):\\(-?[0-9\\.]+\\)$")) x 0
          then 
            let clumpId   = int_of_string (Str.matched_group 1 x) in
            let paramName = Str.matched_group 2 x in
            let start     = float_of_string (Str.matched_group 3 x) in
            let step      = float_of_string (Str.matched_group 4 x) in
            let stop      = float_of_string (Str.matched_group 5 x) in
            let numSteps  = int_of_float (ceil ((stop-.start+.step) /. step)) in
            let arr       = init numSteps (fun i -> 2. ** ((float_of_int i) *. step +. start)) in
              (clumpId, paramName, arr, setP clumpId paramName) :: getOptParams ds xs
          else if Str.string_match (regexp ("^\\([0-9]+\\):\\([^:]+\\):\\(-?[0-9\\.]+\\):\\(-?[0-9\\.]+\\):\\(-?[0-9\\.]+\\)$")) x 0
          then
            let clumpId   = int_of_string (Str.matched_group 1 x) in
            let paramName = Str.matched_group 2 x in
            let start     = float_of_string (Str.matched_group 3 x) in
            let step      = float_of_string (Str.matched_group 4 x) in
            let stop      = float_of_string (Str.matched_group 5 x) in
            let numSteps  = int_of_float (ceil ((stop-.start+.step) /. step)) in
            let arr       = init numSteps (fun i -> (float_of_int i) *. step +. start) in
              (clumpId, paramName, arr, setP clumpId paramName) :: getOptParams ds xs
          else if Str.string_match (regexp "^c:log2:\\(-?[0-9\\.]+\\):\\(-?[0-9\\.]+\\):\\(-?[0-9\\.]+\\)$") x 0
          then 
            let start     = float_of_string (Str.matched_group 1 x) in
            let step      = float_of_string (Str.matched_group 2 x) in
            let stop      = float_of_string (Str.matched_group 3 x) in
            let numSteps  = int_of_float (ceil ((stop-.start+.step) /. step)) in
            let arr       = init numSteps (fun i -> 2. ** ((float_of_int i) *. step +. start)) in
              (-1, "c", arr, fun _ -> ()) :: getOptParams ds xs
              
          else if Str.string_match (regexp "^c:\\(-?[0-9\\.]+\\):\\(-?[0-9\\.]+\\):\\(-?[0-9\\.]+\\)$") x 0
          then 
            let start     = float_of_string (Str.matched_group 1 x) in
            let step      = float_of_string (Str.matched_group 2 x) in
            let stop      = float_of_string (Str.matched_group 3 x) in
            let numSteps  = int_of_float (ceil ((stop-.start+.step) /. step)) in
            let arr       = init numSteps (fun i -> (float_of_int i) *. step +. start) in
              (-1, "c", arr, fun _ -> ()) :: getOptParams ds xs
          else
            getOptParams ds xs
      end

let optModel mname name args =
  ensureData name ;
  let d    = H.find dataEnv name in
  let eta  = orelse float_of_string  (findArg "eta" args)  0.1 in
  let numi = orelse int_of_string    (findArg "numi" args) 1000 in
  let eps  = orelse float_of_string  (findArg "epsilon" args) 0.0001 in
  let c    = orelse float_of_string  (findArg "c" args) 1. in
  let mc   = 
    if List.mem "ova" args then Some (OVA (countClasses d.dataclasses))
    else if List.mem "ava" args then Some (AVA (countClasses d.dataclasses))
    else if List.mem "ovn" args then Some (OVN (countClasses d.dataclasses))
    else if length (countClasses d.dataclasses) > 2 then failwith "need to specify multiclass type (ova,ovn,ava) for mc data"
    else None in

  (* all that is copied form estModel, the next part is for finding params
     to optimize.  in general, for clump i, you can optimize any parameter
     of i (say, beta) by saying one of i:beta:-2:1:3 (will try -2,-1,0,1,2,3)
     or by saying i:beta:log2:-2:1:3 (will try 0.25, 0.5, 1, 2, 4, 8); you
     can, of course, also just optimize c with c:log2:-2:1:2 or ... *)
  let optP = of_list (getOptParams d args) in
  let nP   = length optP in

    Printf.printf "Optimizing:\n" ;
    iter (fun (id,p,_,_) -> 
            if id == -1 
            then Printf.printf "  '%s'\n" p
            else Printf.printf "  '%s' from clump %d\n" p id) optP ;

  let optimizingC = ref false in
  let cArray = ref (-1) in
  let n = length d.dataclasses in
  let err = ref n in

  let optVals = 
    init nP 
      (fun i -> 
         let (id,nam,arr,_) = optP.(i) in 
           if id == -1 && nam = "c" then (optimizingC := true; cArray := i);
           length arr / 2) in

  (* we iteratively hold all but i constant and optimize i *)
  (try
    for const = 0 to length optP - 1 do
      let (id,nam,arr,setP) = optP.(const) in
      let curVal = optVals.(const) in

      if !optimizingC && const = !cArray
      then dprintf "\nCurrently optimizing c\n"
      else dprintf "\nCurrently optimizing %s from clump %d\n" nam id;

      let cval = ref
        (if !optimizingC && const == !cArray
         then arr.(curVal)
         else c) in

      (* set everything else *)
      for i=0 to length optP - 1 do
        if i != const then begin
          let (_,_,arr,setP) = optP.(i) in
          let optV = arr.(optVals.(i)) in
            if !optimizingC && i == !cArray
            then cval := optV
            else setP optV;
        end;
      done;

(*      if const != !cArray then setP arr.(curVal); *)

      let setIt v =
        if const != !cArray then setP arr.(v) else cval := arr.(v) in

      (* optimize the model *)
      let curErr = setIt curVal ; Estimate.crossValidate d eta numi eps !cval mc in
	
	dprintf "\t%s = %g => error rate of %d\n" nam arr.(curVal) curErr;

      let belowErr = 
        if curErr == 0 || curVal == 0 then length d.datapoints
        else ( setIt (curVal-1) ; let e = Estimate.crossValidate d eta numi eps !cval mc in
	       dprintf "\t%s = %g => error rate of %d\n" nam arr.(curVal-1) e ; e ) in

      let aboveErr = 
        if curErr == 0 || belowErr == 0 || curVal == length arr-1 then length d.datapoints
        else ( setIt (curVal+1) ; let e = Estimate.crossValidate d eta numi eps !cval mc in
	       dprintf "\t%s = %g => error rate of %d\n" nam arr.(curVal+1) e ; e ) in

      (* if any of these is zero, we can just stop right now; otherwise,
         we search in the direction pointed to be the gradient (choosing, arbitrarily,
         to go up if there is no difference *)
         
      if curErr   == 0 then ( optVals.(const) <- curVal   ; err := 0 ; raise Break ) ;
      if belowErr == 0 then ( optVals.(const) <- curVal-1 ; err := 0 ; raise Break ) ;
      if aboveErr == 0 then ( optVals.(const) <- curVal+1 ; err := 0 ; raise Break ) ;

      if curErr <= belowErr && curErr <= aboveErr then begin
        optVals.(const) <- curVal ; 
        err := min !err curErr ; 
        Printf.printf "Optimal value at %g (err=%d)\n\n" arr.(curVal) curErr;
      end else begin

        let direction =      if belowErr < aboveErr then -1       else 1         in
        let lastE     = ref (if belowErr < aboveErr then belowErr else aboveErr) in
        let bestPos   = ref (if belowErr < aboveErr then curVal-1 else curVal+1) in
        let curR      = ref (curVal + 2 * direction) in
          
          (try
             while ((!curR >= 0) && (!curR < length arr)) do
               setIt !curR ;
               let e = Estimate.crossValidate d eta numi eps !cval mc in
	         dprintf "\t%s = %g => error rate of %d\n" nam arr.(!curR) e;
                 if e > !lastE then (setIt (!curR-direction) ; raise Break );
                 bestPos := !curR;
	         optVals.(const) <- !curR;
                 lastE   := e;
                 curR := !curR + direction;
             done;
           with
               Break -> ()
             | e -> raise e);

          Printf.printf "Optimal value at %g (err=%d)\n\n" arr.(!bestPos) !lastE;
          optVals.(const) <- !bestPos;
          err := !lastE;
      end;
    done
  with
      Break -> ()
    | e -> raise e) ;
    Printf.printf "All parameters optimized (accuracy = %d/%d = %g%c)\n" (n - !err) n (100. *. float_of_int (n - !err) /. float_of_int n) '%';
    iteri (fun i (id,p,arr,_) -> 
            if id == -1 
            then Printf.printf "  '%s'               \t= %g\n" p    arr.(optVals.(i))
            else Printf.printf "  '%s' from clump %d \t= %g\n" p id arr.(optVals.(i))) optP ;
    Printf.printf "\n";

    (* now, use what we learned to train a model *)
    let cval = if !optimizingC then let (_,_,arr,_) = optP.(!cArray) in arr.(optVals.(!cArray)) else c in
    let model = Estimate.estimateModel "" d None eta numi eps cval mc in
    dprintf "Model %s estimated from %s\n" mname name;
    H.replace modelEnv mname model

let estSig mname name =
  ensureModel mname;
  ensureData name;
  let d = H.find dataEnv name in let m = H.find modelEnv mname in
    match m.multiclass with
        None -> 
          let _ = Printf.printf "Training sigmoid: " in
          let (a,b) = Sigmoid.trainBinarySigmoid m d in
          let _ = Printf.printf ", done!\n" in
            m.binarySigmoid <- Some (a,b)
      | _ -> internal_error "training sigmoid for non-binary class data is not yet implemented"

let saveM name filename =
  ensureModel name;
  BinaryIO.saveModel filename (H.find modelEnv name) ;
  dprintf "Model %s saved to %s\n" name filename

let loadM name filename =
  let m = BinaryIO.loadModel filename in
  H.replace modelEnv name m;
  dprintf "Model %s loaded from to %s\n" name filename


let saveD name filename =
  ensureData name;
  BinaryIO.saveDataSet filename (H.find dataEnv name) ;
  dprintf "Data set %s saved to %s\n" name filename

let loadD name filename =
  let d = BinaryIO.loadDataSet filename in
  H.replace dataEnv name d;
  dprintf "Data set %s loaded from to %s\n" name filename

let normD dataset byD byM =
  ensureData dataset;
  let data = H.find dataEnv dataset in
  match (byD, byM) with
      (None,None) -> ( Normalize.normalizeDataBySelf data ; dprintf "Data '%s' normalized\n" dataset )
    | (Some d,None) -> 
        ( ensureData d ; 
          match (H.find dataEnv d).datanormalize with
              None -> failwith ("dataset '" ^ d ^ "' is not itself normalized")
            | Some a -> Normalize.normalizeDataBy data a ; dprintf "Data '%s' normalized by data '%s'\n" dataset d)
    | (None,Some m) ->
        ( ensureModel m ; 
          match (H.find modelEnv m).modelnormalize with
              None -> failwith ("model '" ^ m ^ "' is not normalized")
            | Some a -> Normalize.normalizeDataBy data a ; dprintf "Data '%s' normalized by model '%s'\n" dataset m)
    | _ -> failwith ""

let who () =
  Printf.printf "Data sets:\n";
  H.iter (fun n d -> Printf.printf "\t%s\t(%d clumps, %d points)\n" n (length d.trainingParams) (length d.datapoints)) dataEnv;
  Printf.printf "\nModels:\n";
  H.iter (fun n m -> Printf.printf "\t%s\t(%d clumps, %d svs)\n" n (length m.modelParams) (length m.supportVectors)) modelEnv;
  Printf.printf "\nVariables:\n";
  H.iter (fun n d -> Printf.printf "\t%s : \t%s\n" n d) valEnv;
  Printf.printf "\n"
  
let clearAll () = H.clear dataEnv ; H.clear modelEnv
let clearData = function
    [] -> H.clear dataEnv
  | l  -> List.iter (H.remove dataEnv) l
let clearModel = function
    [] -> H.clear modelEnv
  | l  -> List.iter (H.remove modelEnv) l

let loadKM filename name clump =
  ensureData name ;
  let d = H.find dataEnv name in
    if clump >= length d.trainingParams then failwith "invalid clump number" else
      d.trainingParams.(clump) <-
        match d.trainingParams.(clump) with
            MatP (w,m) -> MatP (w, DataIO.loadKernelMatrix filename)
          | _ -> failwith "must load kernel matrices into kernel matrix clumps"

let showHelp () = 
  Printf.printf "SVMsequel shell is a machine learning shell for performing estimation\n";
  Printf.printf "(learning), prediction (classification), cross validation and more\n";
  Printf.printf "with support vector machines.  SVMsequel currently lives at:\n";
  Printf.printf "\n";
  Printf.printf "  http://www.isi.edu/~hdaume/SVMsequel/\n";
  Printf.printf "\n";
  Printf.printf "At that URL (and hopefully distributed with this program), you can\n";
  Printf.printf "find full documentation in both postscript and PDF formats\n";
  Printf.printf "(svmsequel.ps and svmsequel.pdf, respectively).  Hopefully that\n";
  Printf.printf "documentation will answer any question you have.  This help is\n";
  Printf.printf "intended more as a quick reference.\n";
  Printf.printf "\n";
  Printf.printf "Available commands:\n";
  Printf.printf "\n";
  Printf.printf "new data <dataset>               creates a new, empty dataset\n" ;
  Printf.printf "\n" ;
  Printf.printf "add <type> clump to              add a new clump to the given dataset <type>\n" ;
  Printf.printf "  <dataset> <args>               in { sparse, dense, dpstring, ststring,\n" ;
  Printf.printf "                                 discrete, matrix }; <args> specifies kernel \n" ;
  Printf.printf "                                 type, eg. 'kernel rbf gamma 10'\n" ;
  Printf.printf "\n" ;
  Printf.printf "modify clump <#> of              modify an existing clump; you cannot change\n" ;
  Printf.printf "  <dataset> to <args>            the type, but you can change the kernel\n" ;
  Printf.printf "                                 parameters (see 'add')\n" ;
  Printf.printf "\n" ;
  Printf.printf "data info <dataset>              show vitals about <dataset>\n" ;
  Printf.printf "\n" ;
  Printf.printf "scan <ftype> file <path>         scan a file of type <ftype> with path <path>;\n" ;
  Printf.printf "                                 <ftype> in { svm, csv, ssv }\n" ;
  Printf.printf "\n" ;
  Printf.printf "load <ftype> file <path>         load a file of type <ftype> (see 'scan') from\n" ;
  Printf.printf "  to <dataset> <columninfo>      path <path> into the (already existing) data\n" ;
  Printf.printf "                                 <dataset>.  <columninfo> specifies how the\n" ;
  Printf.printf "                                 data is interpreted; the format is 'clump <n>\n" ;
  Printf.printf "                                 from <cols>' where <n> is the destination\n" ;
  Printf.printf "                                  clump number and <cols> is the column set\n" ;
  Printf.printf "                                 to put into that clump, eg '3-8,12' or 'all'\n" ;
  Printf.printf "\n" ;
  Printf.printf "load kernel matrix from <path>   load a kernel matrix from <path> to clump\n";
  Printf.printf "  to <dataset> clump <#>         <#> of <dataset>; <path> should be space-\n";
  Printf.printf "                                 separated and either square or upper-triangular\n" ;
  Printf.printf "\n" ;
  Printf.printf "save environment to <path>       save all models, datasets and variables\n";
  Printf.printf "\n" ;
  Printf.printf "load environment from <path>     load a previously saved environment\n";
  Printf.printf "\n" ;
  Printf.printf "normalize data <dataset> <args>  normalize the data in <dataset>;\n" ;
  Printf.printf "                                 if <args> is 'by data <dataset2>', we use\n" ;
  Printf.printf "                                    the normalization from <dataset2>;\n" ;
  Printf.printf "                                 if <args> is 'by model <mname>', we use\n" ;
  Printf.printf "                                    normalization from model <mname>;\n" ;
  Printf.printf "                                 if <args> is empty, we normalize based on\n" ;
  Printf.printf "                                    the given data.\n" ;
  Printf.printf "\n" ;
  Printf.printf "cross-validate <dataset>         prepare <dataset> for cross-validation in\n" ;
  Printf.printf "  <n> folds                      <n> folds; prereq for 'optimize'\n" ;
  Printf.printf "\n" ;
  Printf.printf "estimate model <mname> from      estimate a model called <mname> from the data\n" ;
  Printf.printf "  <dataset> <args>               in <dataset>;  <args> specifies training\n" ;
  Printf.printf "                                 params (c, eta, numi, multi-class type, etc.)\n" ;
  Printf.printf "\n" ;
  Printf.printf "optimize model <mname> from      same as 'estimate', but parameters are\n" ;
  Printf.printf "  <dataset> <args>               specified as sets and a line search is run to\n" ;
  Printf.printf "                                 find good values; eg. 'c:log2:-3:1:3' will\n" ;
  Printf.printf "                                 search for c between 2^-3 and 2^3 with steps\n" ;
  Printf.printf "                                 of 1; '0:gamma:0.5:0.5:2' will search for the\n" ;
  Printf.printf "                                 gamma parameter for clump 0 between 0.5 and 2\n" ;
  Printf.printf "                                 with a step of 0.5\n" ;
  Printf.printf "\n" ;
  Printf.printf "estimate sigmoid for <mname>     train probabilistic parameters (currently only\n" ;
  Printf.printf "  from <dataset>                 for binary classifiers) for model <mname>\n" ;
  Printf.printf "                                 based on its performance on dataset <dataset>\n";
  Printf.printf "\n";
  Printf.printf "model info <mname>               display vitals for model <mname>\n" ;
  Printf.printf "\n" ;
  Printf.printf "predict <dataset> with           predict values for <dataset> using model\n" ;
  Printf.printf "  <mname> <args>                 <mname>; in general this just produces\n" ;
  Printf.printf "                                 accuracies: to get values put in <args>\n" ;
  Printf.printf "                                 'output <filename>' or 'output -' for stdout\n" ;
  Printf.printf "\n" ;
  Printf.printf "run <script> <args>              run svmsequel script <script> with <args>\n" ;
  Printf.printf "                                 (does not quit at end, <args> overwrites,\n" ;
  Printf.printf "                                 temporarily, our arguments)\n" ;
  Printf.printf "\n" ;
  Printf.printf "save model <mname> to <path>     save model <mname> to <path>\n" ;
  Printf.printf "\n" ;
  Printf.printf "load model <mname> from <path>   load the model\n" ;
  Printf.printf "\n" ;
  Printf.printf "save data <dataset> to <path>    save dataset <dataset> to <path>\n" ;
  Printf.printf "\n" ;
  Printf.printf "load data <dataset> from <path>  load the data back\n" ;
  Printf.printf "\n" ;
  Printf.printf "require <#> <usage>              for scripts, fail if less than <#> arguments\n" ;
  Printf.printf "                                 are given, displaying <usage> as an error\n" ;
  Printf.printf "\n" ;
  Printf.printf "require <arg> <usage>            for scripts, fail if less argument <arg> is\n" ;
  Printf.printf "                                 not given, displaying <usage> as an error\n" ;
  Printf.printf "\n" ;
  Printf.printf "who                              list active datasets, models and variables\n" ;
  Printf.printf "\n" ;
  Printf.printf "clear <args>                     if <args> is empty, delte all datasets and \n" ;
  Printf.printf "                                 model; if <args> is 'data ...', clear datasets\n" ;
  Printf.printf "                                 ...; if <args> is 'model ...', clear models...\n" ;
  Printf.printf "\n" ;
  Printf.printf "echo <args>                      echo a string\n" ;
  Printf.printf "\n" ;
  Printf.printf "!<command>                       issue <command> to shell\n";
  Printf.printf "\n" ;
  Printf.printf "help/h/?                         show this help\n" ;
  Printf.printf "\n" ;
  Printf.printf "quit/exit/q                      quit\n"

exception Not_enough_arguments

let addVar x v = H.replace valEnv (String.sub x 1 (String.length x - 1)) v
let addQVar x xs =
  let ifExists = takeWhile (fun s -> not (s = "|")) xs in
  let ifNot    = drop (List.length ifExists+1) xs in
    if H.mem valEnv (String.sub x 1 (String.length x - 1))
    then H.replace valEnv (String.sub x 1 (String.length x - 1)) (String.concat " " ifExists)
    else H.add     valEnv (String.sub x 1 (String.length x - 1)) (String.concat " " ifNot)

let rec runScript lineNum x2 =
  if string_match (regexp "^[ \t]*$") x2 0 || string_match (regexp "^#") x2 0 
  then ()
  else if string_match (regexp "^!.") x2 0
  then let _ = Sys.command (String.sub x2 1 (String.length x2-1)) in ()
  else let x = String.concat "" (replaceVariables (explode x2)) in begin
    (match split (regexp "[ \t]+") x with
         [] -> ()
       | ["new" ; "data" ; name] -> newData name
       | ("add"::typ::"clump"::"to"::name::args) -> addClump name typ args
       | ("modify"::"clump"::clumpNum::"of"::name::"to"::args) -> modClump name (int_of_string clumpNum) args
       | ("data"::"info"::name::[]) -> showDataInfo name
       | ("scan"::filetype::"file"::filepath::[]) -> scanFile filepath filetype
       | ("load"::"model"::name::"from"::filename::[]) -> loadM name filename
       | ("load"::"data"::name::"from"::filename::[]) -> loadD name filename
       | ("load"::"kernel"::"matrix"::"from"::filename::"to"::name::"clump"::n::[]) -> loadKM filename name (int_of_string n)
       | ("load"::"environment"::"from"::filename::[]) -> BinaryIO.loadEnvironment filename dataEnv modelEnv valEnv
       | ("load"::filetype::"file"::filepath::"to"::name::columninfo) -> loadFile name filepath filetype columninfo
       | ("cross-validate"::name::n::"folds"::[]) -> setupXV name (int_of_string n)
       | ("estimate"::"model"::modelname::"from"::name::args) -> estModel modelname name args
       | ("optimize"::"model"::modelname::"from"::name::args) -> optModel modelname name args
       | ("model"::"info"::name::[]) -> showModelInfo name
       | ("predict"::name::"with"::modelname::args) -> pred name modelname args
       | ("run"::scriptname::args) -> runScriptFile scriptname false args
       | ("save"::"model"::name::"to"::filename::[]) -> saveM name filename
       | ("save"::"environment"::"to"::filename::[])   -> BinaryIO.saveEnvironment filename dataEnv modelEnv valEnv
       | ("save"::"data"::name::"to"::filename::[]) -> saveD name filename
       | ("normalize"::"data"::name::[]) -> normD name None None
       | ("normalize"::"data"::name::"by"::"data"::name2::[]) -> normD name (Some name2) None
       | ("normalize"::"data"::name::"by"::"model"::mname::[]) -> normD name None (Some mname) 
       | ("require"::n::usage) -> 
           if H.mem valEnv n then ()
           else (Printf.printf "%s\n" (String.concat " " usage) ; raise Not_enough_arguments)
       | ("echo"::args) -> Printf.printf "%s\n" (String.concat " " args)
       | ["who"] -> who ()
       | ["clear"] -> clearAll ()
       | ("clear"::"data"::args)  -> clearData args
       | ("clear"::"model"::args) -> clearModel args

       | ("estimate"::"sigmoid"::"for"::mname::"from"::name::[]) -> estSig mname name

       | ["help"] | ["h"] | ["?"] -> showHelp ()

       | (x::"="::xs) when String.length x > 0 && String.get x 0 == '@' -> addVar x (String.concat "" xs)
       | (x::"=?"::xs) when String.length x > 0 && String.get x 0 == '@' -> addQVar x xs

       | _ -> failwith ("error on line " ^ string_of_int lineNum ^ ", unknown command (type ? for help):\n  " ^ x));
  end
and runScriptFile scriptname useCmdLneArgs args =
  let oldValEnv = 
    if useCmdLneArgs then valEnv
    else
      let v = H.copy valEnv in
      let _ = H.clear valEnv in
      let _ = List.fold_left (fun i v -> H.replace valEnv (string_of_int i) v ; i+1) 1 args in
        v in
  let h = open_in scriptname in
  let line = ref 0 in
    try
      while (true) do
	let lne = input_line h in
	  if lne = "quit" || lne = "exit" then raise End_of_file;
	  line := !line+1;
	  try
	    runScript !line lne
	  with
	      Failure s ->
		Printf.printf "ERROR IN SCRIPT %s [%d]: %s\n" scriptname !line s
	    | e -> raise e
      done
    with
	End_of_file -> 
          ( if not useCmdLneArgs then
              ( H.clear valEnv;
                H.iter (fun k v -> H.replace valEnv k v) oldValEnv ) )
      | e -> raise e


let _ = if not versiononly then dprintf "  ____               ____  ___  ___      ___\n / __/|| // /\\/\\    / __/ /    /  / /  //   /   SVMsequel shell version %s\n_\\_ \\ ||// /    \\  _\\_ \\ /--  /  / /  //-- /    www.isi.edu/~hdaume/SVMsequel/\n\\___/ |_/ / /\\/\\ \\ \\___//___ /__\\ /__//___/___  %s\n\n" version (if isNone inputfile then "'?' for help, 'q' to quit" else "")

let _ = Random.self_init ()

let _ =
  if versiononly then Printf.printf "SVMsequel shell version %s\n" version 
  else if usage then printusage ()
  else 
    match inputfile with
        Some fn -> runScriptFile fn true []
      | None    -> 
          let linenum = ref 0 in
            try
              while true do
                let _ = print_string "> " in
                let lne = read_line () in
                  if lne = "quit" || lne = "exit" || lne = "q" then raise End_of_file;
                  linenum := !linenum + 1;
                  try
                    runScript !linenum lne
                  with 
                      Failure s ->
                        Printf.printf "ERROR [%d]: %s\n" !linenum s;
                    | e -> raise e;
              done;
            with
                End_of_file -> ()
              | e -> raise e;
          
