open Array
open Printf

type 'a ptr = { ptr : 'a ref option ; ptr_id : int }

let globalPtrId = ref 1

type suffixTreeNode =
    { mutable sons : suffixTreeNode ptr;
      mutable right_sibling : suffixTreeNode ptr;
      mutable left_sibling  : suffixTreeNode ptr;
      mutable father        : suffixTreeNode ptr;
      mutable suffix_link   : suffixTreeNode ptr;
      mutable path_position : int;
      mutable edge_label_start : int;
      mutable edge_label_end   : int;
      (* these are filled in last, so they are unique and zero-based *)
      mutable identifier : int;  
      mutable num_leaves : int;
      mutable string_length : int
    }

type suffixTree =
    { mutable e : int;
      mutable tree_string : int array;
      mutable root : suffixTreeNode ptr;
      mutable num_nodes : int
    }

type skipType = Skip | NoSkip
type rule2Type = NewSon | Split
type lastPosType = LastCharInEdge | OtherChar

type suffixTreePath =
    { mutable p_begin : int;
      mutable p_end   : int
    }

type suffixTreePos =
    { mutable node : suffixTreeNode ptr;
      mutable edge_pos : int
    }

type nodeP = suffixTreeNode ptr
type posP  = suffixTreePos  ptr

let eqPtr (x : 'a ptr) (y : 'a ptr) : bool = x.ptr_id == y.ptr_id

let nullP : 'a ptr = { ptr = None ; ptr_id = 0 }
let newPtr (a : 'a) : 'a ptr = let p = !globalPtrId in let _ = incr globalPtrId in { ptr = Some (ref a) ; ptr_id = p }

let isNull p = match p.ptr with None -> true | _ -> false
let deref p = match p.ptr with None -> failwith "dereferencing of null ptr" | Some r -> !r

let create_node (father : nodeP) st en pos : nodeP =
  newPtr { sons = nullP ; 
           right_sibling = nullP;
           left_sibling  = nullP;
           suffix_link   = nullP;
           father        = father;
           path_position = pos;
           edge_label_start = st;
           edge_label_end   = en;
           identifier = -1;
           num_leaves = -1;
           string_length = -1;
         }

let last_son (node : nodeP) : nodeP =
  let rec go_right nd =
    if isNull (deref nd).right_sibling
    then nd
    else go_right (deref nd).right_sibling in
  if isNull node then nullP
  else if isNull (deref node).sons then nullP
  else go_right (deref node).sons

let find_son (tree : suffixTree) (node : nodeP) (character : int) : nodeP =
  let rec find_son' nd =
    if isNull nd then nd
    else if tree.tree_string.((deref nd).edge_label_start-1) == character then nd
    else find_son' (deref nd).right_sibling in

    find_son' (deref node).sons

let get_node_label_end (tree : suffixTree) (node : nodeP) : int =
  if isNull (deref node).sons then tree.e
  else (deref node).edge_label_end

let get_node_label_length (tree : suffixTree) (node : nodeP) : int =
  get_node_label_end tree node - (deref node).edge_label_start + 1

let is_last_char_in_edge (tree : suffixTree) (node : nodeP) (ep : int) : bool =
  ep+1 == get_node_label_length tree node


let printTree (tree : suffixTree) =

  let rec printNode nd depth =
    if depth > 0 then (
      for i=1 to depth-1 do Printf.printf "|"; done;
      Printf.printf "+[%d] " (deref nd).identifier;
      for i=(deref nd).edge_label_start to get_node_label_end tree nd do
        if tree.tree_string.(i-1) == -1
        then Printf.printf "$ "
        else Printf.printf "%d " tree.tree_string.(i-1);
      done;
      Printf.printf "\n"
    );
    printSons (deref nd).sons (depth+1)

  and printSons nd depth =
    if isNull nd then () else
      ( printNode nd depth;
        printSons (deref nd).right_sibling depth ) in

    printNode tree.root 0
      

let connect_siblings (left_sib : nodeP) (right_sib : nodeP) : unit =
  if not (isNull left_sib ) then (deref left_sib ).right_sibling <- right_sib;
  if not (isNull right_sib) then (deref right_sib).left_sibling  <- left_sib

let apply_extension_rule_2 (node : nodeP) (elb : int) (ele : int) (pp : int) (ep : int) (typ : rule2Type) : nodeP =
  match typ with
      NewSon -> begin
        let new_leaf = create_node node elb ele pp in
          connect_siblings (last_son node) new_leaf;
          new_leaf
      end

    | Split -> begin
        let new_internal = create_node (deref node).father (deref node).edge_label_start ((deref node).edge_label_start + ep) (deref node).path_position in

          (deref node).edge_label_start <- (deref node).edge_label_start + ep + 1;

        let new_leaf = create_node new_internal elb ele pp in
          
          connect_siblings (deref node).left_sibling new_internal;
          connect_siblings new_internal (deref node).right_sibling;
          (deref node).left_sibling <- nullP;
          
          if eqPtr ((deref (deref new_internal).father).sons) node
          then (deref (deref new_internal).father).sons <- new_internal;

          (deref new_internal).sons <- node;
          (deref node).father <- new_internal;
          connect_siblings node new_leaf;
          new_internal
      end

let trace_single_edge (tree : suffixTree) (node : nodeP) (str : suffixTreePath) (ep : int ref) (chars_found : int ref) (typ : skipType) (search_done : bool ref) : nodeP =
  search_done := true;
  ep := 0;
  let cont_node = find_son tree node tree.tree_string.(str.p_begin-1) in
    if isNull cont_node
    then ( ep := get_node_label_length tree node - 1 ; chars_found := 0 ; node ) else
      
  let node = cont_node in
  let length = ref (get_node_label_length tree node) in
  let str_len = ref (str.p_end - str.p_begin + 1) in

    match typ with
        Skip -> begin
          if !length <= !str_len 
          then ( chars_found := !length ;
                 ep          := !length - 1;
                 if (!length < !str_len) then search_done := false )
          else ( chars_found := !str_len;
                 ep          := !str_len-1 );

          node
        end

      | NoSkip -> begin
          if !str_len < !length then length := !str_len;

          ep := 1; chars_found := 1;
          let isdone = ref false in
            while ((not !isdone) && (!ep < !length)) do
              if tree.tree_string.((deref node).edge_label_start + !ep-1) != tree.tree_string.(str.p_begin + !ep-1)
              then (ep := !ep - 1; isdone := true)
              else (
                incr chars_found;
                incr ep
              )
            done;
            
            if not !isdone 
            then ( ep := !ep - 1;
                   if !chars_found < !str_len then search_done := false );

            node
        end

let trace_string (tree : suffixTree) (node : nodeP) (str : suffixTreePath) (ep : int ref) (chars_found : int ref) (typ : skipType) : nodeP =
  let str' = { p_begin = str.p_begin ; p_end = str.p_end } in
  let search_done = ref false in
  let new_node = ref node in
    chars_found := 0;
    while (not !search_done) do
      let edge_chars_found = ref 0 in
        ep := 0;
        new_node := trace_single_edge tree !new_node str ep edge_chars_found typ search_done;
        str.p_begin <-  str.p_begin + !edge_chars_found;
        chars_found := !chars_found + !edge_chars_found;
    done;
    str.p_begin <- str'.p_begin; str.p_end <- str'.p_end;
    !new_node

let follow_suffix_link (tree : suffixTree) (pos : posP) : unit =
  let gama = { p_begin = -1 ; p_end = -1 } in
  let chars_found = ref 0 in
  if eqPtr tree.root (deref pos).node then () 
  else if isNull (deref (deref pos).node).suffix_link ||
          not (is_last_char_in_edge tree (deref pos).node (deref pos).edge_pos)
  then (
    if eqPtr (deref (deref pos).node).father tree.root
    then (deref pos).node <- tree.root
    else ( gama.p_begin <- (deref (deref pos).node).edge_label_start;
           gama.p_end   <- (deref (deref pos).node).edge_label_start + (deref pos).edge_pos;
           (deref pos).node <- (deref (deref (deref pos).node).father).suffix_link;
           let ep = ref (deref pos).edge_pos in
           (deref pos).node <- trace_string tree (deref pos).node gama ep chars_found Skip;
           (deref pos).edge_pos <- !ep )
  ) else (
    (deref pos).node <- (deref (deref pos).node).suffix_link;
    (deref pos).edge_pos <- get_node_label_length tree (deref pos).node - 1;
  )

let create_suffix_link (node : nodeP) (link : nodeP) : unit =
  (deref node).suffix_link <- link

let suffixless : nodeP ref = ref nullP

let sea (tree : suffixTree) (pos : posP) (str : suffixTreePath) (rule_applied : int ref) (after_rule_3 : bool) : unit =
  let chars_found = ref 0 in
  let path_pos    = str.p_begin in
  let tmp = ref nullP in
    
    if not after_rule_3 then follow_suffix_link tree pos;

    if eqPtr (deref pos).node tree.root
    then (
      let ep = ref (deref pos).edge_pos in
        (deref pos).node <- trace_string tree tree.root str ep chars_found NoSkip;
        (deref pos).edge_pos <- !ep
    ) else (
      str.p_begin <- str.p_end;
      chars_found := 0;

      if is_last_char_in_edge tree (deref pos).node (deref pos).edge_pos
      then (
        tmp := find_son tree (deref pos).node tree.tree_string.(str.p_end-1);
        if not (isNull !tmp) then (
          (deref pos).node <- !tmp;
          (deref pos).edge_pos <- 0;
          chars_found := 1
        )
      ) else (
        if tree.tree_string.((deref (deref pos).node).edge_label_start + (deref pos).edge_pos + 1-1) == 
           tree.tree_string.(str.p_end-1)
        then ( (deref pos).edge_pos <- (deref pos).edge_pos + 1;
               chars_found := 1 )
      )
    );

    if !chars_found == str.p_end - str.p_begin + 1
    then (
      rule_applied := 3;
      if not (isNull !suffixless)
      then ( create_suffix_link !suffixless (deref (deref pos).node).father; 
             suffixless := nullP );
      ()
    ) else if is_last_char_in_edge tree (deref pos).node (deref pos).edge_pos ||
              eqPtr (deref pos).node tree.root
    then (
      if not (isNull (deref (deref pos).node).sons)
      then (
        let _ = apply_extension_rule_2 (deref pos).node (str.p_begin + !chars_found) str.p_end path_pos 0 NewSon in
        rule_applied := 2;
        if not (isNull !suffixless)
        then ( create_suffix_link !suffixless (deref pos).node;
               suffixless := nullP )
      )
    ) else (
      tmp := apply_extension_rule_2 (deref pos).node (str.p_begin + !chars_found) str.p_end path_pos (deref pos).edge_pos Split;

      if not (isNull !suffixless) then create_suffix_link !suffixless !tmp;

      if get_node_label_length tree !tmp == 1 && eqPtr (deref !tmp).father tree.root
      then ( (deref !tmp).suffix_link <- tree.root; suffixless := nullP )
      else suffixless := !tmp;

      (deref pos).node <- !tmp;
      rule_applied := 2;
    )

let spa (tree : suffixTree) (pos : posP) (phase : int) (extension : int ref) (repeated_extension : bool ref) : unit =
  let rule_applied = ref 0 in
  let str = { p_begin = -1; p_end = -1 } in
  let isdone = ref false in

    tree.e <- phase+1;
    while ((not !isdone) && (!extension <= phase+1)) do
      str.p_begin <- !extension;
      str.p_end   <- phase+1;

      sea tree pos str rule_applied !repeated_extension;

      if !rule_applied == 3
      then ( repeated_extension := true ; isdone := true )
      else ( repeated_extension := false; incr extension );
    done;
    ()

let setIdentifiers (tree : suffixTree) : unit =
  let globalIdentifier = ref 0 in
  let rec setId nd =
    if isNull nd || (deref nd).identifier >= 0 then () else begin
      (deref nd).identifier <- !globalIdentifier;
      incr globalIdentifier;
      setId (deref nd).sons;
      setId (deref nd).right_sibling;
    end in
    setId tree.root;
    tree.num_nodes <- !globalIdentifier;
    ()

let setNumLeaves (tree : suffixTree) : unit =
  let rec setIt nd =
    if isNull (deref nd).sons 
    then (deref nd).num_leaves <- 1
    else (
      let n = ref (deref nd).sons in
      let s = ref 0 in
        while (not (isNull !n)) do
          setIt !n;
          s := !s + (deref !n).num_leaves;
          n := (deref !n).right_sibling;
        done;
        (deref nd).num_leaves <- !s
    ) in
    setIt tree.root

let setStringLength (tree : suffixTree) : unit =
  let rec setIt nd len =
    let my_len = len + (get_node_label_end tree nd) - (deref nd).edge_label_start + 1 in
    let n      = ref (deref nd).sons in
      (deref nd).string_length <- my_len;
      while not (isNull !n) do
        setIt !n my_len;
        n := (deref !n).right_sibling;
      done;
      () in
    setIt tree.root 0

(* str should be padded with -1 *)
let createTree (str : int array) =
  let len = length str in
  let r   = create_node nullP 0 0 0 in
  let extension = ref 2 in
  let repeated_extension = ref false in
  let phase = ref 1 in

  let tree = { e = 0; tree_string = str; root = r ; num_nodes = -1 } in

  let pos  = newPtr { node = r ; edge_pos = 0 } in

    (deref r).sons <- create_node tree.root 1 len 1;
    suffixless := nullP;
    while (!phase < len) do
      spa tree pos !phase extension repeated_extension;
      incr phase;
    done;

    setIdentifiers  tree;
    setNumLeaves    tree;
    setStringLength tree;
    tree
(*
(* let t = createTree [| 1 ; 2 ; 1 ; 2 ; 3 ; -1 |] *)
let str = init (length Sys.argv) (fun i -> if i < length Sys.argv-1 then int_of_string (Sys.argv.(i+1)) else -1)
let t = createTree str
let _ = printTree t
*)

