open SuffixTree
open Array
open Printf

type weightType = 
    ExpDecay of float | ConstWeight | BoundedRange of int

type streeData = { tree_ref : suffixTree; tree_kvc : float array }

let kv_exp_decay lambda gamma tau =
  (lambda ** (0.-.float_of_int gamma) -. lambda ** (0.-.float_of_int tau)) /. (lambda -. 1.)

let kv_const_weight gamma tau = float_of_int tau -. float_of_int gamma

let kv_bounded_range n gamma tau = float_of_int (max 0 ((min tau n) - gamma))


let weightTypeToFn = function
    ExpDecay lambda -> kv_exp_decay lambda
  | ConstWeight -> kv_const_weight
  | BoundedRange n -> kv_bounded_range n

let computeKernelValCache weightFn t =
  let kernelVal = make t.num_nodes 0. in
  let rec computeVal parVal w parLen =
    let newLen = parLen + (get_node_label_end t w) - (deref w).edge_label_start + 1 in
    let myVal  = parVal +. (float_of_int (deref w).num_leaves) *. weightFn parLen newLen in
    let nd     = ref (deref w).sons in
      kernelVal.((deref w).identifier) <- myVal;
      while (not (isNull !nd)) do
        computeVal myVal !nd newLen;
        nd := (deref !nd).right_sibling;
      done in

  let nd = ref (deref t.root).sons in
    while (not (isNull !nd)) do
      computeVal 0. !nd 0;
      nd := (deref !nd).right_sibling;
    done;
    kernelVal


let calcMatchingStatistics t x =
  let n  = length x in
  let v  = make (n-1) 0 in
  let c  = make (n-1) nullP in
  let c' = make (n-1) nullP in

  let rec getMatchLength i j end_pos len = 
    if i >= length x-1
    then (true, j > end_pos, len)
    else if j > end_pos
    then (false, true, len)
    else if x.(i) == t.tree_string.(j-1)
    then getMatchLength (i+1) (j+1) end_pos (len+1)
    else (false, false, len) in

  let rec followString i parent nd acc_len =
    if isNull nd 
    then (acc_len, parent, nd) 
    else 
      match getMatchLength i (deref nd).edge_label_start (get_node_label_end t nd) 0 with
           (_,_,0)           -> followString i parent (deref nd).right_sibling acc_len
         | (true ,true ,len) -> (len+acc_len, parent, nd)
         | (true ,false,len) -> (len+acc_len, parent, nd)
         | (false,false,len) -> (len+acc_len, parent, nd)
         | (false,true ,len) -> 
             if isNull (deref nd).sons 
             then (len+acc_len, parent, nd)
             else followString (i+len) nd (deref nd).sons (acc_len+len) in

  let rec findMS i node =
    let (vv,cv',cv) = followString i node (deref node).sons 0 in
      v.(i) <- vv;
      c.(i) <- cv;
      c'.(i) <- cv';

      if i < length x - 2
      then
        if not (isNull (deref cv').sons) (* internal node *)
        then
          if isNull (deref cv').suffix_link
          then findMS (i+1) cv' (* t.root *)
          else findMS (i+1) (deref cv').suffix_link
        else 
          if eqPtr t.root (deref cv').father
          then findMS (i+1) t.root
          else
            if isNull (deref (deref cv').father).suffix_link
            then findMS (i+1) (deref cv').father (* t.root *)
            else findMS (i+1) (deref (deref cv').father).suffix_link
      else
        () in
    findMS 0 t.root;
    (v,c,c')

let st_string_kernel weightFn t kvc x =
  let (va,ca,ca') = calcMatchingStatistics t x in
  let s = ref 0. in
    for i=0 to length x - 2 do
      let v  = va.(i) in
      let c  = ca.(i) in
      let c' = ca'.(i) in
        if not (isNull c') then (
(*          printf "id=%d kvc=%g\n" (deref c').identifier kvc.((deref c').identifier); *)
          s := !s +. kvc.((deref c').identifier);
          if not (isNull c ) then (
            let gamma = (deref c').string_length - 1 in
            let tau   = (deref c ).string_length - 1 in
(*              printf "num_leaves=%d gamma=%d tau=%d\n" (deref c).num_leaves gamma tau; *)
              s := !s +. (float_of_int (deref c).num_leaves) *. weightFn gamma tau
          )
        );
    done;
    !s
(*
let t = createTree [|1;2;1;2;3;-1|]
let kvc = computeKernelValCache (kv_exp_decay 0.75) t
let _ = printf "%g\n" (st_string_kernel (kv_exp_decay 0.75) t kvc [|4;-1|])
*)

let k x y =
  let t = createTree x in
  let kvc = computeKernelValCache (kv_exp_decay 0.75) t in
    st_string_kernel (kv_exp_decay 0.75) t kvc y
